import numpy as np
try:
    from macer.externals.dynaphopy_bundled.c.displacements import atomic_displacements
except (ImportError, ModuleNotFoundError):
    # Fallback to pure python implementation if C extension is missing or fails (e.g. NumPy version mismatch)
    def atomic_displacements(trajectory, position, cell):
        # Implementation similar to DynaPhoPy's C-extension fallback
        # trajectory: (n_steps, 3), position: (3,), cell: (3, 3)
        diff = trajectory - position
        inv_cell = np.linalg.inv(cell)
        frac_diff = np.dot(diff, inv_cell)
        frac_diff -= np.round(frac_diff)
        return np.dot(frac_diff, cell)

import os


class Dynamics:

    def __init__(self,
                 structure=None,
                 trajectory=None,
                 scaled_trajectory=None,
                 velocity=None,
                 energy=None,
                 time=None,
                 supercell=None,
                 memmap=False):

        self._time = time
        self._trajectory = trajectory
        self._scaled_trajectory = scaled_trajectory
        self._energy = energy
        self._velocity = velocity
        self._supercell = supercell
        self._memmap=memmap

        self._time_step_average = None
        self._velocity_mass_average = None
        self._relative_trajectory = None
        self._supercell_matrix = None
        self._number_of_atoms = None
        self._mean_displacement_matrix = None

        if structure is not None:
            self._structure = structure

            if trajectory is not None:
                self._trajectory = trajectory

        else:
            print('Warning: Initialization without structure')
            self._structure = None

        #Read environtment variables
        try:
            self._temp_directory = os.environ["DYNAPHOPY_TEMPDIR"]
            if os.path.isdir(self._temp_directory):
                self._temp_directory = self._temp_directory + '/'
            else:
                self._temp_directory = ''
        except KeyError:
            self._temp_directory = ''

    def __del__(self):
        #Clean all temporal files from memmap
        if self._memmap:
            for mapped_array in [self._velocity, self._trajectory, self._relative_trajectory, self._velocity_mass_average]:
                try:
                    filename = mapped_array.filename
                except AttributeError:
                    continue
                del mapped_array
                os.remove(filename)

# A bit messy, has to be fixed
    def crop_trajectory(self, last_steps):

        if last_steps is None or last_steps <= 0:
            return

        if self._trajectory is not None:
            self._trajectory = self._trajectory[-last_steps:, :, :]
            self._relative_trajectory = None
            self._scaled_trajectory = None

        if self._relative_trajectory is not None:
            self._relative_trajectory = self._relative_trajectory[-last_steps:, :, :]
            self._trajectory = None
            self._scaled_trajectory = None

        if self._scaled_trajectory is not None:
            self._scaled_trajectory = self._scaled_trajectory[-last_steps:, :, :]
            self._trajectory = None
            self._relative_trajectory = None

        if self._energy is not None:
            self._energy = self._energy[-last_steps:]
        if self._time is not None:
            self._time = self._time[-last_steps:]

        if last_steps > self.velocity.shape[0]:
            print("Warning: specified step number larger than available")

        self.velocity = self.velocity[-last_steps:, :, :]

        self._velocity_mass_average = None

        if self._memmap:
            filename = self._relative_trajectory.filename
            self._relative_trajectory = None
            try:
                os.remove(filename)
            except:
                pass
        else:
            self._relative_trajectory = None

        # print("Using {0} steps".format(self.velocity.shape[0]))

    def get_number_of_atoms(self):
        if self._number_of_atoms is None:
            self._number_of_atoms = self.structure.get_number_of_atoms()*np.prod(self.get_supercell_matrix())
        return self._number_of_atoms

    def set_time(self, time):
        self._time = time

    def get_time(self):
        return self._time

    #def set_supercell(self, supercell):
    #    self._supercell = supercell

    def get_supercell(self):
        return self._supercell

    def get_energy(self):
        return self._energy

    def get_time_step_average(self):

        if not self._time_step_average :
            self._time_step_average = 0
            for i in range(len(self.get_time()) - 1):
                self._time_step_average += (self.get_time()[i+1] - self.get_time()[i])/(len(self.get_time()) - 1)
   #         self._time_step_average /= (self.get_time().shape[0]-1)
        self._time_step_average = np.round(self._time_step_average, decimals=8)

        return self._time_step_average

    def set_structure(self, structure):
        self._structure = structure

    def get_velocity_mass_average(self):
        if self._velocity_mass_average is None:
            if self._memmap:
                self._velocity_mass_average = np.memmap(self._temp_directory+'velocity_mass.{0}'.format(os.getpid()),
                                                        dtype='complex', mode='w+', shape=self.velocity.shape)
            else:
                self._velocity_mass_average = np.empty_like(self.velocity)

            supercell = self.get_supercell_matrix()
            for i in range(self.get_number_of_atoms()):
                self._velocity_mass_average[:, i, :] = (self.velocity[:, i, :] *
                                                        np.sqrt(self.structure.get_masses(supercell=supercell)[i]))

        return self._velocity_mass_average

    def get_relative_trajectory(self):
        if self._relative_trajectory is None:

            trajectory = self.trajectory

            supercell = self.get_supercell()
            number_of_atoms = self.trajectory.shape[1]
            supercell_matrix = self.get_supercell_matrix()
            position = self.structure.get_positions(supercell=supercell_matrix)


            if self._memmap:
                normalized_trajectory = np.memmap(self._temp_directory+'r_trajectory.{0}'.format(os.getpid()),
                                                  dtype='complex', mode='w+', shape=trajectory.shape)
            else:
                normalized_trajectory = self.trajectory.copy()

            for i in range(number_of_atoms):
                normalized_trajectory[:, i, :] = atomic_displacements(trajectory[:, i, :], position[i], supercell)

            self._relative_trajectory = normalized_trajectory
        return self._relative_trajectory

    def get_supercell_matrix(self, tolerance=0.01):

        def parameters2(h):
            a = np.linalg.norm(h[:,0])
            b = np.linalg.norm(h[:,1])
            c = np.linalg.norm(h[:,2])
            return [a, b, c]

        def parameters(h):
            #return [np.linalg.norm(h[i, :]) for i in range(h.shape[1])]
            return np.linalg.norm(h, axis=1)

        if self._supercell_matrix is None:
            supercell_matrix_real = np.divide(parameters(self.get_supercell()), parameters(self.structure.get_cell()))
            self._supercell_matrix = np.around(supercell_matrix_real).astype("int")

            if abs(sum(self._supercell_matrix - supercell_matrix_real)/np.linalg.norm(supercell_matrix_real)) > tolerance:
                print(abs(sum(self._supercell_matrix - supercell_matrix_real) / np.linalg.norm(supercell_matrix_real)))
                print('Warning! Defined unit cell and MD supercell do not match!')
                print('Cell size relation is not integer: {0}'.format(supercell_matrix_real))
                exit()

            print('MD cell size relation: {0}'.format(self._supercell_matrix))

        return self._supercell_matrix

    def get_mean_displacement_matrix(self, use_average_positions=False):

        if self._mean_displacement_matrix is None:

            atom_type = self.structure.get_atom_type_index()
            number_atom_primitive_equivalent = np.unique(atom_type, return_counts=True)[1]

            supercell = self.get_supercell_matrix()
            atom_type_index = self.structure.get_atom_type_index(supercell=supercell)
            number_of_atom_types = self.structure.get_number_of_atom_types()
            number_of_dimensions = self.structure.get_number_of_dimensions()
            displacements = self.get_relative_trajectory()
            number_of_data = displacements.shape[0]

            # Correct the atom positions by position average
            if use_average_positions:
                position_difference = np.average(displacements, axis=0)
            else:
                position_difference = np.zeros_like(self.structure.get_positions(supercell=supercell))

            number_of_equivalent_atoms = np.prod(supercell)

            mean_displacement_matrix = np.zeros((number_of_atom_types, number_of_dimensions, number_of_dimensions))

            for i in range(displacements.shape[1]):
                primtive_normalization = number_atom_primitive_equivalent[atom_type_index[i]]
                mean_displacement_matrix[atom_type_index[i], :, :] += np.dot(np.conj(displacements[:, i, :]).T,
                                                                             displacements[:, i, :] - position_difference[i]
                                                                             ).real / primtive_normalization

            self._mean_displacement_matrix = mean_displacement_matrix / (number_of_equivalent_atoms * number_of_data)

        return self._mean_displacement_matrix

    def average_positions(self, number_of_samples=None, to_unit_cell=False):

        supercell = self.get_supercell_matrix()
        number_of_dimensions = self.structure.get_number_of_dimensions()

        cell = self.get_supercell()
        number_of_atoms = self.trajectory.shape[1]
        positions = self.structure.get_positions(supercell=supercell)

        normalized_trajectory = self.get_relative_trajectory()

        if number_of_samples:
            length = normalized_trajectory.shape[0]
            if length < number_of_samples:
                number_of_samples = length
            normalized_trajectory = normalized_trajectory
            samples = np.random.random_integers(length, size=(number_of_samples,))-1
            normalized_trajectory = normalized_trajectory[samples, :]

        averaged_positions = np.average(normalized_trajectory, axis=0)

        # Average respect to unit cell
        if to_unit_cell:

            cell = self.structure.get_cell()
            number_of_atoms = self.structure.get_number_of_atoms()
            positions = self.structure.get_positions()

            index_type_unitcell = self.structure.get_atom_type_index()
            index_type = self.structure.get_atom_type_index(supercell=supercell)

            number_of_atom_types = self.structure.get_number_of_atom_types()
            normalization = np.prod(supercell)

            averaged_unit_cell = np.zeros((number_of_atom_types, number_of_dimensions), dtype=complex)

            for i, coordinates  in enumerate(averaged_positions):
                averaged_unit_cell[index_type[i], :] += coordinates/normalization

            averaged_positions = []
            for i in range(positions.shape[0]):
                averaged_positions.append(averaged_unit_cell[index_type_unitcell[i]])
            averaged_positions = np.array(averaged_positions)

        # Average respect to unit cell
        averaged_positions += positions

        for j in range(number_of_atoms):

            difference_matrix = np.around(np.dot(np.linalg.inv(cell.T),
                                                 averaged_positions[j, :] - 0.5 * np.dot(np.ones((number_of_dimensions)), cell)),
                                                 decimals=0)
            averaged_positions[j, :] -= np.dot(difference_matrix, cell)

        return averaged_positions

    # Properties
    @property
    def structure(self):
        return self._structure

    @property
    def trajectory(self):
        if self._trajectory is None:
            if self._scaled_trajectory is not None:
                self._trajectory = np.dot(self._scaled_trajectory, self.get_supercell())
            else:
                print('No trajectory loaded')
                exit()

        return self._trajectory

    @property
    def velocity(self):
        if self._velocity is None:
            print('No velocity provided! calculating it from coordinates...')
            if self._memmap:
                self._velocity = np.memmap(self._temp_directory+'velocity.{0}'.format(os.getpid()),
                                           dtype='complex',
                                           mode='w+',
                                           shape=self.get_relative_trajectory().shape)
            else:
                self._velocity = np.zeros_like(self.get_relative_trajectory(), dtype=complex)
            for i in range(self.get_number_of_atoms()):
                for j in range(self.structure.get_number_of_dimensions()):
                    self._velocity[:, i, j] = np.gradient(self.get_relative_trajectory()[:, i, j],
                                                          self.get_time_step_average())

        return self._velocity

    @velocity.setter
    def velocity(self,velocity):
        self._velocity = velocity


"""Base class of force constants solvers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np

from macer.externals.symfc_bundled.basis_sets import FCBasisSetBase


class FCSolverBase(ABC):
    """Abstract base class of force constants solvers."""

    def __init__(
        self,
        basis_set: Union[FCBasisSetBase, Sequence[FCBasisSetBase]],
        use_mkl: bool = False,
        log_level: int = 0,
    ):
        """Init method."""
        self._basis_set = basis_set
        self._use_mkl = use_mkl
        self._log_level = log_level
        if isinstance(self._basis_set, Sequence):
            _basis_set = self._basis_set[0]
        else:
            _basis_set = basis_set
        assert isinstance(_basis_set, FCBasisSetBase)
        self._natom = _basis_set.translation_permutations.shape[1]
        self._coefs: Optional[Union[np.ndarray, Sequence[np.ndarray]]] = None

    @property
    def coefs(self) -> Optional[Union[np.ndarray, Sequence[np.ndarray]]]:
        """Return coefficients of force constants with respect to basis set."""
        return self._coefs

    @property
    @abstractmethod
    def full_fc(self) -> Optional[np.ndarray]:
        """Return full force constants."""
        pass

    @property
    @abstractmethod
    def compact_fc(self) -> Optional[np.ndarray]:
        """Return compact force constants."""
        pass

    @abstractmethod
    def solve(self):
        """Solve coefficients of basis set from displacements and forces."""
        pass



###########################################################################################
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################
from __future__ import annotations

import argparse
import ast
import logging
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Tuple, Union

import ase.data
import ase.io
import numpy as np
import torch

from macer.externals.mace_bundled.calculators import MACECalculator, mace_mp

try:
    import fpsample  # type: ignore
except ImportError:
    pass


class FilteringType(Enum):
    NONE = "none"
    COMBINATIONS = "combinations"
    EXCLUSIVE = "exclusive"
    INCLUSIVE = "inclusive"


class SubselectType(Enum):
    FPS = "fps"
    RANDOM = "random"


@dataclass
class SelectionSettings:
    configs_pt: str
    output: str
    configs_ft: str | None = None
    atomic_numbers: List[int] | None = None
    num_samples: int | None = None
    subselect: SubselectType = SubselectType.FPS
    model: str = "small"
    descriptors: str | None = None
    device: str = "cpu"
    default_dtype: str = "float64"
    head_pt: str | None = None
    head_ft: str | None = None
    filtering_type: FilteringType = FilteringType.COMBINATIONS
    weight_ft: float = 1.0
    weight_pt: float = 1.0
    allow_random_padding: bool = True
    seed: int = 42


def str_to_list(s: str) -> List[int]:
    assert isinstance(s, str), "Input must be a string"
    return ast.literal_eval(s)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--configs_pt",
        help="path to XYZ configurations for the pretraining",
        required=True,
    )
    parser.add_argument(
        "--configs_ft",
        help="path or list of paths to XYZ configurations for the finetuning",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--num_samples",
        help="number of samples to select for the pretraining",
        type=int,
        required=False,
        default=None,
    )
    parser.add_argument(
        "--subselect",
        help="method to subselect the configurations of the pretraining set",
        type=SubselectType,
        choices=list(SubselectType),
        default=SubselectType.FPS,
    )
    parser.add_argument(
        "--model", help="path to model", default="small", required=False
    )
    parser.add_argument("--output", help="output path", required=True)
    parser.add_argument(
        "--descriptors", help="path to descriptors", required=False, default=None
    )
    parser.add_argument(
        "--device",
        help="select device",
        type=str,
        choices=["cpu", "cuda"],
        default="cpu",
    )
    parser.add_argument(
        "--default_dtype",
        help="set default dtype",
        type=str,
        choices=["float32", "float64"],
        default="float64",
    )
    parser.add_argument(
        "--head_pt",
        help="level of head for the pretraining set",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--head_ft",
        help="level of head for the finetuning set",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--filtering_type",
        help="filtering type",
        type=FilteringType,
        choices=list(FilteringType),
        default=FilteringType.NONE,
    )
    parser.add_argument(
        "--weight_ft",
        help="weight for the finetuning set",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--weight_pt",
        help="weight for the pretraining set",
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "--atomic_numbers",
        help="list of atomic numbers to filter the configurations",
        type=str_to_list,
        default=None,
    )
    parser.add_argument(
        "--disallow_random_padding",
        help="do not allow random padding of the configurations to match the number of samples",
        action="store_false",
        dest="allow_random_padding",
    )
    parser.add_argument("--seed", help="random seed", type=int, default=42)
    return parser.parse_args()


def calculate_descriptors(atoms: List[ase.Atoms], calc: MACECalculator) -> None:
    logging.info("Calculating descriptors")
    for mol in atoms:
        descriptors = calc.get_descriptors(mol.copy(), invariants_only=True)
        # average descriptors over atoms for each element
        descriptors_dict = {
            element: np.mean(descriptors[mol.symbols == element], axis=0)
            for element in np.unique(mol.symbols)
        }
        mol.info["mace_descriptors"] = descriptors_dict


def filter_atoms(
    atoms: ase.Atoms,
    element_subset: List[str],
    filtering_type: FilteringType = FilteringType.COMBINATIONS,
) -> bool:
    """
    Filters atoms based on the provided filtering type and element subset.

    Parameters:
    atoms (ase.Atoms): The atoms object to filter.
    element_subset (list): The list of elements to consider during filtering.
    filtering_type (FilteringType): The type of filtering to apply.
        Can be one of the following `FilteringType` enum members:
          - `FilteringType.NONE`: No filtering is applied.
          - `FilteringType.COMBINATIONS`: Return true if `atoms` is composed of combinations of elements in the subset, false otherwise. I.e. does not require all of the specified elements to be present.
          - `FilteringType.EXCLUSIVE`: Return true if `atoms` contains *only* elements in the subset, false otherwise.
          - `FilteringType.INCLUSIVE`: Return true if `atoms` contains all elements in the subset, false otherwise. I.e. allows additional elements.

    Returns:
    bool: True if the atoms pass the filter, False otherwise.
    """
    if filtering_type == FilteringType.NONE:
        return True
    if filtering_type == FilteringType.COMBINATIONS:
        atom_symbols = np.unique(atoms.symbols)
        return all(
            x in element_subset for x in atom_symbols
        )  # atoms must *only* contain elements in the subset
    if filtering_type == FilteringType.EXCLUSIVE:
        atom_symbols = set(list(atoms.symbols))
        return atom_symbols == set(element_subset)
    if filtering_type == FilteringType.INCLUSIVE:
        atom_symbols = np.unique(atoms.symbols)
        return all(
            x in atom_symbols for x in element_subset
        )  # atoms must *at least* contain elements in the subset
    raise ValueError(
        f"Filtering type {filtering_type} not recognised. Must be one of {list(FilteringType)}."
    )


class FPS:
    def __init__(self, atoms_list: List[ase.Atoms], n_samples: int):
        self.n_samples = n_samples
        self.atoms_list = atoms_list
        self.species = np.unique([x.symbol for atoms in atoms_list for x in atoms])  # type: ignore
        self.species_dict = {x: i for i, x in enumerate(self.species)}
        # start from a random configuration
        self.list_index = [np.random.randint(0, len(atoms_list))]
        self.assemble_descriptors()

    def run(
        self,
    ) -> List[int]:
        """
        Run the farthest point sampling algorithm.
        """
        descriptor_dataset_reshaped = (
            self.descriptors_dataset.reshape(  # pylint: disable=E1121
                (len(self.atoms_list), -1)
            )
        )
        logging.info(f"{descriptor_dataset_reshaped.shape}")
        logging.info(f"n_samples: {self.n_samples}")
        self.list_index = fpsample.fps_npdu_kdtree_sampling(
            descriptor_dataset_reshaped,
            self.n_samples,
        )
        return self.list_index

    def assemble_descriptors(self) -> None:
        """
        Assemble the descriptors for all the configurations.
        """
        self.descriptors_dataset: np.ndarray = 10e10 * np.ones(
            (
                len(self.atoms_list),
                len(self.species),
                len(list(self.atoms_list[0].info["mace_descriptors"].values())[0]),
            ),
            dtype=np.float32,
        ).astype(np.float32)

        for i, atoms in enumerate(self.atoms_list):
            descriptors = atoms.info["mace_descriptors"]
            for z in descriptors:
                self.descriptors_dataset[i, self.species_dict[z]] = np.array(
                    descriptors[z]
                ).astype(np.float32)


def _load_calc(
    model: str, device: str, default_dtype: str, subselect: SubselectType
) -> Union[MACECalculator, None]:
    if subselect == SubselectType.RANDOM:
        return None
    if model in ["small", "medium", "large"]:
        calc = mace_mp(model, device=device, default_dtype=default_dtype)
    else:
        calc = MACECalculator(
            model_paths=model,
            device=device,
            default_dtype=default_dtype,
        )
    return calc


def _get_finetuning_elements(
    atoms: List[ase.Atoms], atomic_numbers: List[int] | None
) -> List[str]:
    if atoms:
        logging.debug(
            "Using elements from the finetuning configurations for filtering."
        )
        species = np.unique([x.symbol for atoms in atoms for x in atoms]).tolist()  # type: ignore
    elif atomic_numbers is not None and atomic_numbers:
        logging.debug("Using the supplied atomic numbers for filtering.")
        species = [ase.data.chemical_symbols[z] for z in atomic_numbers]
    else:
        species = []
    return species


def _read_finetuning_configs(
    configs_ft: Union[str, list[str], None],
) -> List[ase.Atoms]:
    if isinstance(configs_ft, str):
        path = configs_ft
        return ase.io.read(path, index=":")  # type: ignore
    if isinstance(configs_ft, list):
        assert all(isinstance(x, str) for x in configs_ft)
        atoms_list_ft = []
        for path in configs_ft:
            atoms_list_ft += ase.io.read(path, index=":")
        return atoms_list_ft
    if configs_ft is None:
        return []
    raise ValueError(f"Invalid type for configs_ft: {type(configs_ft)}")


def _filter_pretraining_data(
    atoms: list[ase.Atoms],
    filtering_type: FilteringType,
    all_species_ft: List[str],
) -> Tuple[List[ase.Atoms], List[ase.Atoms], list[bool]]:
    logging.info(
        "Filtering configurations based on the finetuning set, "
        f"filtering type: {filtering_type}, elements: {all_species_ft}"
    )
    passes_filter = [filter_atoms(x, all_species_ft, filtering_type) for x in atoms]
    assert len(passes_filter) == len(atoms), "Filtering failed"
    filtered_atoms = [x for x, passes in zip(atoms, passes_filter) if passes]
    remaining_atoms = [x for x, passes in zip(atoms, passes_filter) if not passes]
    return filtered_atoms, remaining_atoms, passes_filter


def _get_random_configs(
    num_samples: int,
    atoms: List[ase.Atoms],
) -> list[ase.Atoms]:
    if num_samples > len(atoms):
        raise ValueError(
            f"Requested more samples ({num_samples}) than available in the remaining set ({len(atoms)})"
        )
    indices = np.random.choice(list(range(len(atoms))), num_samples, replace=False)
    return [atoms[i] for i in indices]


def _load_descriptors(
    atoms: List[ase.Atoms],
    passes_filter: List[bool],
    descriptors_path: str | None,
    calc: MACECalculator | None,
    full_data_length: int,
) -> None:
    if descriptors_path is not None:
        logging.info(f"Loading descriptors from {descriptors_path}")
        descriptors = np.load(descriptors_path, allow_pickle=True)
        assert sum(passes_filter) == len(atoms)
        if len(descriptors) != full_data_length:
            raise ValueError(
                f"Length of the descriptors ({len(descriptors)}) does not match the length of the data ({full_data_length})"
                "Please provide descriptors for all configurations"
            )
        required_descriptors = [
            descriptors[i] for i, passes in enumerate(passes_filter) if passes
        ]
        for i, atoms_ in enumerate(atoms):
            atoms_.info["mace_descriptors"] = required_descriptors[i]
    else:
        logging.info("Calculating descriptors")
        if calc is None:
            raise ValueError("MACECalculator must be provided to calculate descriptors")
        calculate_descriptors(atoms, calc)


def _maybe_save_descriptors(
    atoms: List[ase.Atoms],
    output_path: str,
) -> None:
    """
    Save the descriptors if they are present in the atoms objects.
    Also, delete the descriptors from the atoms objects.
    """
    if all("mace_descriptors" in x.info for x in atoms):
        output_path = Path(output_path)
        descriptor_save_path = output_path.parent / (
            output_path.stem + "_descriptors.npy"
        )
        logging.info(f"Saving descriptors at {descriptor_save_path}")
        descriptors_list = [x.info["mace_descriptors"] for x in atoms]
        np.save(descriptor_save_path, descriptors_list, allow_pickle=True)
        for x in atoms:
            del x.info["mace_descriptors"]


def _maybe_fps(atoms: List[ase.Atoms], num_samples: int) -> List[ase.Atoms]:
    try:
        fps_pt = FPS(atoms, num_samples)
        idx_pt = fps_pt.run()
        logging.info(f"Selected {len(idx_pt)} configurations")
        return [atoms[i] for i in idx_pt]
    except Exception as e:  # pylint: disable=W0703
        logging.error(f"FPS failed, selecting random configurations instead: {e}")
        return _get_random_configs(num_samples, atoms)


def _subsample_data(
    filtered_atoms: List[ase.Atoms],
    remaining_atoms: List[ase.Atoms],
    passes_filter: List[bool],
    num_samples: int | None,
    subselect: SubselectType,
    descriptors_path: str | None,
    allow_random_padding: bool,
    calc: MACECalculator | None,
) -> List[ase.Atoms]:
    if num_samples is None or num_samples == len(filtered_atoms):
        logging.info(
            f"No subsampling, keeping all {len(filtered_atoms)} filtered configurations"
        )
        return filtered_atoms
    if num_samples > len(filtered_atoms) and allow_random_padding:
        num_sample_randomly = num_samples - len(filtered_atoms)
        logging.info(
            f"Number of configurations after filtering {len(filtered_atoms)} "
            f"is less than the number of samples {num_samples}, "
            f"selecting {num_sample_randomly} random configurations for the rest."
        )
        return filtered_atoms + _get_random_configs(
            num_sample_randomly, remaining_atoms
        )
    if num_samples == 0:
        raise ValueError("Number of samples must be greater than 0")
    if subselect == SubselectType.FPS:
        _load_descriptors(
            filtered_atoms,
            passes_filter,
            descriptors_path,
            calc,
            full_data_length=len(filtered_atoms) + len(remaining_atoms),
        )
        logging.info(
            f"Selecting {num_samples} configurations out of {len(filtered_atoms)} using Farthest Point Sampling"
        )
        return _maybe_fps(filtered_atoms, num_samples)
    if subselect == SubselectType.RANDOM:
        logging.info(
            f"Subselecting {num_samples} from filtered {len(filtered_atoms)} using random sampling"
        )
        return _get_random_configs(num_samples, filtered_atoms)
    raise ValueError(f"Invalid subselect type: {subselect}")


def _write_metadata(
    atoms: list[ase.Atoms], pretrained: bool, config_weight: float, head: str | None
) -> None:
    for a in atoms:
        a.info["pretrained"] = pretrained
        a.info["config_weight"] = config_weight
        if head is not None:
            a.info["head"] = head


def select_samples(
    settings: SelectionSettings,
) -> None:
    np.random.seed(settings.seed)
    torch.manual_seed(settings.seed)
    calc = _load_calc(
        settings.model, settings.device, settings.default_dtype, settings.subselect
    )
    atoms_list_ft = _read_finetuning_configs(settings.configs_ft)
    all_species_ft = _get_finetuning_elements(atoms_list_ft, settings.atomic_numbers)

    if settings.filtering_type is not FilteringType.NONE and not all_species_ft:
        raise ValueError(
            "Filtering types other than NONE require elements for filtering. They can be specified via the `--atomic_numbers` flag."
        )

    atoms_list_pt: list[ase.Atoms] = ase.io.read(settings.configs_pt, index=":")  # type: ignore
    filtered_pt_atoms, remaining_atoms, passes_filter = _filter_pretraining_data(
        atoms_list_pt, settings.filtering_type, all_species_ft
    )

    subsampled_atoms = _subsample_data(
        filtered_pt_atoms,
        remaining_atoms,
        passes_filter,
        settings.num_samples,
        settings.subselect,
        settings.descriptors,
        settings.allow_random_padding,
        calc,
    )
    if ase.io.formats.filetype(settings.output, read=False) != "extxyz":
        raise ValueError(
            f"filename '{settings.output}' does no have "
            "suffix compatible with extxyz format"
        )
    _maybe_save_descriptors(subsampled_atoms, settings.output)

    _write_metadata(
        subsampled_atoms,
        pretrained=True,
        config_weight=settings.weight_pt,
        head=settings.head_pt,
    )
    _write_metadata(
        atoms_list_ft,
        pretrained=False,
        config_weight=settings.weight_ft,
        head=settings.head_ft,
    )

    logging.info("Saving the selected configurations")
    ase.io.write(settings.output, subsampled_atoms)

    logging.info("Saving a combined XYZ file")
    atoms_fps_pt_ft = subsampled_atoms + atoms_list_ft

    output = Path(settings.output)
    ase.io.write(
        output.parent / (output.stem + "_combined" + output.suffix),
        atoms_fps_pt_ft,
    )


def main():
    args = parse_args()
    settings = SelectionSettings(**vars(args))
    select_samples(settings)


if __name__ == "__main__":
    main()



"""
__author__ = "Hiroki Koiso"
__copyright__ = "Copyright 2023, Nakajima group"
__version__ = "2.1"
__maintainer__ = "Hiroki Koiso"
__email__ = "koiso.h.aa@m.titech.ac.jp"
__status__ = "Development"
__date__ = "July 22, 2023"
"""

import plotly.graph_objs as go
import plotly.offline as pyo
import pandas as pd
import numpy as np
import argparse, yaml

### Parser structure ###
desc = """
This program plots the average Grüneisen parameter at each q-point. 
Before using this program, you need to make gruneisen.yaml obtained as following:
phonopy-gruneisen orig plus minus --dim="2 2 2" --mesh="20 20 20" -c POSCAR --nomeshsym
"""
parser = argparse.ArgumentParser(description=desc)
parser.add_argument("--yaml", "-yaml", dest="name_of_yaml", default="gruneisen.yaml", type=str,
                    help="Input a gruneisen.yaml file.")
parser.add_argument("--color-bar-min", "-cmin", dest="cmin", default=-5, type=float,
                    help="Maximum value of the colorbar.")
parser.add_argument("--color-bar-max", "-cmax", dest="cmax", default=5, type=float,
                    help="Minimum value of the colorbar")
parser.add_argument("--cutoff-freq", "-cut_f", dest="cut_f", default=0.0, type=float,
                    help="the mode-Gruneisen parameters below this cutoff frequency are not used.")
parser.add_argument("--original-position", "-ori-posi", dest="original_position", action="store_true",
                    help="Do not fix q-positions.")
parser.add_argument("--fill-BZ", "-f-BZ", dest="f_BZ", action="store_true",
                    help="Fill in the first Brillouin zone.")
parser.add_argument("--relative-coordinate", "-r-c", dest="plot_rc", action="store_true",
                    help="Plot in relative coordinate.")
args = parser.parse_args()

yaml_file = args.name_of_yaml
cmin = args.cmin
cmax = args.cmax
cut_f = args.cut_f
fix_position = not args.original_position
f_BZ = args.f_BZ
plot_rc = args.plot_rc


### The below function was made by Qijing Zheng. ###
### http://staff.ustc.edu.cn/~zqj/posts/howto-plot-brillouin-zone/ ###
def get_brillouin_zone_3d(cell):
    """ Generate the Brillouin Zone of a given cell.
    The BZ is the Wigner-Seitz cell of the reciprocal lattice, which can be constructed by Voronoi decompositionto the reciprocal lattice. 
    A Voronoi diagram is a subdivision of the space into the nearest neighborhoods of a given set of points. 

    https://en.wikipedia.org/wiki/Wigner%E2%80%93Seitz_cell
    https://docs.scipy.org/doc/scipy/reference/tutorial/spatial.html#voronoi-diagrams

    Parameters
    ----------
    cell : ndarray, shape=(3, 3), raw vectors [b1, b2, b3]^T
        reciprocal lattice vector  

    Returns
    ----------
    vertices : tuple
        vertices of BZ
    ridges : list of list
        ridges of BZ
    faces : list of list
        faces of BZ

    """
    cell = np.asarray(cell, dtype=float)
    assert cell.shape == (3, 3)

    px, py, pz = np.tensordot(cell, np.mgrid[-1:2, -1:2, -1:2], axes=[0, 0])
    points = np.c_[px.ravel(), py.ravel(), pz.ravel()]
    from scipy.spatial import Voronoi
    vor = Voronoi(points)

    bz_facets = []
    bz_ridges = []
    bz_vertices = []

    for pid, rid in zip(vor.ridge_points, vor.ridge_vertices):
        # WHY 13 ????
        # The Voronoi ridges/facets are perpendicular to the lines drawn between the
        # input points. The 14th input point is [0, 0, 0].
        if(pid[0] == 13 or pid[1] == 13):
            bz_ridges.append(vor.vertices[np.r_[rid, [rid[0]]]])
            bz_facets.append(vor.vertices[rid])
            bz_vertices += rid

    bz_vertices = list(set(bz_vertices))

    return vor.vertices[bz_vertices], bz_ridges, bz_facets


def q_position_fixer(q_position, fix_position):
    """ Fix q-points to be within half of the first Brillouin zone. """
    if fix_position == True:
        if q_position[2] <= -0.5:
            q_position = q_position * (-1)
    else:
        pass
    return q_position


def get_average_gruneisen_Brillouin_zone(yaml_file, cut_f, fix_position):
    """ Calculate average Gruneisen parameter at each q-point.

    Parameters
    ----------
    yaml_file : str
        file name of gruneisen.yaml
    cut_f : float
        cut-off frequency
    fix_position : bool
        whether fix q-points to be within half of the first Brillouin zone or do not.

    Returns
    ----------
    list of plot data : ndarray
        shape=((num of q-positions), 7)
    reciprocal_lattice : ndarray, shape=(3, 3)
        raw vectors [b1, b2, b3]^T

    """
    with open(yaml_file, 'r') as file:
        config = yaml.safe_load(file)
    nqpoint = config['nqpoint']
    natom = config['natom']
    nband = 3 * natom
    rec_cell = np.array(config['reciprocal_lattice'])

    data_array = []
    for q in range(nqpoint):
        q_position = q_position_fixer(np.array(config['phonon'][q]['q-position']), fix_position)
        q_position_ac = np.dot(q_position, rec_cell)
        gamma_list =[]
        for band in range(nband):
            gruneisen = config['phonon'][q]['band'][band]['gruneisen']
            frequency = config['phonon'][q]['band'][band]['frequency']
            ### Remove phonon mode below cutoff frequency. ###
            if frequency > cut_f:
                gamma_list.append(gruneisen)
            else:
                pass
        gruneisen_average = np.mean(gamma_list)
        data = [q_position[0], q_position[1], q_position[2], q_position_ac[0], q_position_ac[1], q_position_ac[2],  gruneisen_average]
        data_array.append(data)
    return np.array(data_array), rec_cell


data_array, rec_cell = get_average_gruneisen_Brillouin_zone(yaml_file, cut_f, fix_position)
if f_BZ:
    add_data = data_array.copy()
    add_data[:, list(range(0, 6))] *= (-1)
    data_array = np.concatenate([data_array, add_data])

""" Save data """
with open('data_average_gruneisen_BZ.csv', "w") as fp:
    fp.write('qx_rc,qy_rc,qz_rc,qx_ac,qy_ac,qz_ac,gruneisen_average\n')
    np.savetxt(fp, data_array, fmt='%f', delimiter=',')



""" plot """
df_plot = pd.read_csv('data_average_gruneisen_BZ.csv')

def axis_setting(label, plot_rc):
    axis_setting = dict(
                    showticklabels=plot_rc,
                    title=dict(text=label,
                    font=dict(family='Arial',size=30,color='black')),
                    tickfont=dict(family='Arial',size=16,color='black')
    )
    return axis_setting

if plot_rc:
    qx=df_plot['qx_rc']
    qy=df_plot['qy_rc']
    qz=df_plot['qz_rc']
else: 
    qx=df_plot['qx_ac']
    qy=df_plot['qy_ac']
    qz=df_plot['qz_ac']

colorbar_layout=dict(title=dict(text='&#947;<sub>q</sub>',
                                 font=dict(family='Times New Roman',size=35,color='black')),
                                 tickfont=dict(family='Arial',size=24,color='black'),
                                 len=0.9,
                                 ticks="outside",
                                 tickwidth=2
)
marker_setting = dict(
        size=5,
        color=df_plot['gruneisen_average'],
        #colorscale='Picnic',
        colorscale=[[0.0, 'rgb(0, 0, 255)'], [(0-cmin)/(cmax-cmin), 'rgb(255, 255, 255)'], [1.0, 'rgb(255, 0, 0)']],
        cmin=cmin,
        cmax=cmax,
        opacity=1.0,
        colorbar=colorbar_layout
)
trace = go.Scatter3d(
    x=qx,
    y=qy,
    z=qz,
    customdata=np.stack((df_plot['qx_rc'],df_plot['qy_rc'],df_plot['qz_rc'],df_plot['gruneisen_average']), axis=-1),
    mode='markers',
    hovertemplate='q-position: (%{customdata[0]}, %{customdata[1]}, %{customdata[2]})<br>&#947;<sub>q</sub> average: %{customdata[3]}',
    marker=marker_setting
)

layout = go.Layout(
    margin=dict(l=0, r=0, b=0, t=0),
    showlegend=False,
    scene=dict(
        xaxis=axis_setting('<i>k<sub>x</sub></i>', plot_rc),
        yaxis=axis_setting('<i>k<sub>y</sub></i>', plot_rc),
        zaxis=axis_setting('<i>k<sub>z</sub></i>', plot_rc)
    )
)

fig = go.Figure(data=trace, layout=layout)

if not plot_rc:
    v, e, f = get_brillouin_zone_3d(rec_cell)
    
    # Base
    basis_clrs = ['red', 'green', 'blue']
    basis_labs = ['<i>b<sub>1</sub></i>', '<i>b<sub>2</sub></i>', '<i>b<sub>3</sub></i>']
    for ii, basis in enumerate(rec_cell):
        bx, by, bz = basis
        fig.add_trace(
                go.Scatter3d(
                    x=[0, bx],
                    y=[0, by],
                    z=[0, bz],
                    opacity=0.8,
                    hoverinfo='skip',
                    mode='lines+text',
                    line=dict(
                        color=basis_clrs[ii],
                        width=6,
                    ),
                    text=['', basis_labs[ii]],
                    textfont=dict(color=basis_clrs[ii], size=30),
                )
            )
    # Edige
    for edge in e:
        fig.add_trace(
                go.Scatter3d(
                    x=edge[:, 0],
                    y=edge[:, 1],
                    z=edge[:, 2],
                    opacity=0.8,
                    hoverinfo='skip',
                    mode='lines',
                    line=dict(
                        color='black',
                        width=5,
                    ),
                )
            )
#    # Vartices
#    v_rc = np.dot(v, np.linalg.inv(rec_cell))
#    fig.add_trace(
#            go.Scatter3d(
#                x=v[:, 0],
#                y=v[:, 1],
#                z=v[:, 2],
#                customdata=v_rc,
#                hovertemplate='q-position: (%{customdata[0]:.2f}, %{customdata[1]:.2f}, %{customdata[2]:.2f})',
#                opacity=1,
#                mode='markers',
#                marker=dict(
#                    color='black',
#                    size=3,
#                ),
#            )
#        )

fig.update_layout(
    showlegend=False,
    scene=dict(
        xaxis_visible=False,
        yaxis_visible=False,
        zaxis_visible=False),
    )


pyo.plot(fig, filename=f'average_gruneisen_BZ.html')



# -*- coding: utf-8 -*-
"""
This module is adapted from a script originally by ADI, modified by Gemini.
It extracts phonon eigenvectors from a phonopy `band.yaml` file
and generates VESTA files to visualize the phonon modes as arrows.
It now includes functionality to generate the base VESTA file from a POSCAR
using the same logic as example/vesta-translation/poscar2vesta.py.
"""

import numpy as np
import sys
import os
import yaml
from pathlib import Path
import math

#<editor-fold desc="poscar2vesta Conversion Logic">
def _normalize(v):
    norm = math.sqrt(sum(x*x for x in v))
    return [x/norm for x in v] if norm > 0 else v

def _dot(v1, v2):
    return sum(x*y for x, y in zip(v1, v2))

def _length(v):
    return math.sqrt(sum(x*x for x in v))

def _lattice_params(basis):
    a_vec, b_vec, c_vec = basis
    a = _length(a_vec)
    b = _length(b_vec)
    c = _length(c_vec)
    try:
        alpha = math.degrees(math.acos(round(_dot(b_vec, c_vec) / (b * c), 10)))
        beta = math.degrees(math.acos(round(_dot(a_vec, c_vec) / (a * c), 10)))
        gamma = math.degrees(math.acos(round(_dot(a_vec, b_vec) / (a * b), 10)))
    except (ValueError, ZeroDivisionError):
        alpha, beta, gamma = 90.0, 90.0, 90.0
    return a, b, c, alpha, beta, gamma

def _mat_mul_vec(m, v):
    return [
        m[0][0]*v[0] + m[0][1]*v[1] + m[0][2]*v[2],
        m[1][0]*v[0] + m[1][1]*v[1] + m[1][2]*v[2],
        m[2][0]*v[0] + m[2][1]*v[1] + m[2][2]*v[2]
    ]

def _invert_3x3(m):
    det = m[0][0] * (m[1][1] * m[2][2] - m[2][1] * m[1][2]) - \
          m[0][1] * (m[1][0] * m[2][2] - m[1][2] * m[2][0]) + \
          m[0][2] * (m[1][0] * m[2][1] - m[1][1] * m[2][0])
    if abs(det) < 1e-9: return [[0]*3 for _ in range(3)]
    invDet = 1.0 / det
    minv = [[0]*3 for _ in range(3)]
    minv[0][0] = (m[1][1] * m[2][2] - m[2][1] * m[1][2]) * invDet
    minv[0][1] = (m[0][2] * m[2][1] - m[0][1] * m[2][2]) * invDet
    minv[0][2] = (m[0][1] * m[1][2] - m[0][2] * m[1][1]) * invDet
    minv[1][0] = (m[1][2] * m[2][0] - m[1][0] * m[2][2]) * invDet
    minv[1][1] = (m[0][0] * m[2][2] - m[0][2] * m[2][0]) * invDet
    minv[1][2] = (m[1][0] * m[0][2] - m[0][0] * m[1][2]) * invDet
    minv[2][0] = (m[1][0] * m[2][1] - m[2][0] * m[1][1]) * invDet
    minv[2][1] = (m[2][0] * m[0][1] - m[0][0] * m[2][1]) * invDet
    minv[2][2] = (m[0][0] * m[1][1] - m[1][0] * m[0][1]) * invDet
    return minv

def _read_poscar_for_vesta(filepath):
    with open(filepath, 'r') as f:
        lines = [l.strip() for l in f.readlines() if l.strip()]
    title, scale = lines[0], float(lines[1])
    lattice = [[float(x) * scale for x in lines[i].split()] for i in range(2, 5)]
    
    line5_parts = lines[5].split()
    try:
        counts = [int(x) for x in line5_parts]
        elements = [f"E{i+1}" for i in range(len(counts))]
        coord_line_idx = 6
    except ValueError:
        elements, counts = line5_parts, [int(x) for x in lines[6].split()]
        coord_line_idx = 7
    
    atom_types = [el for el, count in zip(elements, counts) for _ in range(count)]
    
    # Handle optional "Selective dynamics" (Fixed by Gemini)
    if lines[coord_line_idx].strip().lower().startswith('s'):
        coord_line_idx += 1

    coord_type = lines[coord_line_idx].lower()
    coord_start_idx = coord_line_idx + 1
    
    atoms = []
    element_counters = {el: 0 for el in elements}
    
    inv_lattice = None
    if 'cart' in coord_type or 'k' in coord_type:
        inv_lattice = _invert_3x3(np.array(lattice).T)

    for i in range(sum(counts)):
        parts = lines[coord_start_idx + i].split()
        coords = [float(x) for x in parts[:3]]
        
        if inv_lattice:
            coords = _mat_mul_vec(inv_lattice, coords)
        
        # Keep coords within [0, 1) if desired, but VESTA handles >1
        coords = [c % 1.0 for c in coords]
        
        el = atom_types[i]
        element_counters[el] += 1
        atoms.append({'element': el, 'coords': coords, 'label': f"{el}{element_counters[el]}"})

    return title, lattice, atoms, list(elements)

def _generate_vesta_content(title, lattice, atoms, unique_elements):
    """
    Generates the full VESTA file content with standard headers, colors, and footer.
    Matches the style of the user provided CORRECTED file.
    """
    a, b, c, alpha, beta, gamma = _lattice_params(lattice)

    # Header
    content = [
        "#VESTA_FORMAT_VERSION 3.5.4\n\n",
        "CRYSTAL\n\n",
        f"TITLE\n{title}\n\n",
        "GROUP\n1 1 P 1\n",
        "SYMOP\n 0.000000  0.000000  0.000000  1  0  0   0  1  0   0  0  1   1\n -1.0 -1.0 -1.0  0 0 0  0 0 0  0 0 0\n",
        "TRANM 0\n 0.000000  0.000000  0.000000  1  0  0   0  1  0   0  0  1\n",
        "LTRANSL\n -1\n 0.000000  0.000000  0.000000  0.000000  0.000000  0.000000\n",
        "LORIENT\n -1   0   0   0   0\n 1.000000  0.000000  0.000000  1.000000  0.000000  0.000000\n 0.000000  0.000000  1.000000  0.000000  0.000000  1.000000\n",
        "LMATRIX\n 1.000000  0.000000  0.000000  0.000000\n 0.000000  1.000000  0.000000  0.000000\n 0.000000  0.000000  1.000000  0.000000\n 0.000000  0.000000  0.000000  1.000000\n 0.000000  0.000000  0.000000\n",
        "CELLP\n",
        f"  {a:.6f}   {b:.6f}   {c:.6f}  {alpha:.6f}  {beta:.6f}  {gamma:.6f}\n",
        f"  0.000000   0.000000   0.000000   0.000000   0.000000   0.000000\n"
    ]

    # STRUC
    content.append("STRUC\n")
    for i, atom in enumerate(atoms):
        el, lbl, (x, y, z) = atom['element'], atom['label'], atom['coords']
        # Adjusted formatting to match CORRECTED: {i+1} {el}        {lbl}
        content.append(f"  {i+1} {el:<9} {lbl:<4} 1.0000   {x:.6f}   {y:.6f}   {z:.6f}    1a       1\n")
        content.append("                            0.000000   0.000000   0.000000  0.00\n")
    content.append("  0 0 0 0 0 0 0\n")

    # THERI (Dummy)
    content.append(f"THERI {len(atoms)}\n")
    for i, atom in enumerate(atoms):
        content.append(f"  {i+1}        {atom['label']} -0.000000\n")
    content.append("  0 0 0\n")
    
    # SHAPE & BOUND (Dummy defaults)
    content.append("SHAPE\n  0       0       0       0   0.000000  0   192   192   192   192\n")
    content.append("BOUND\n       0        1         0        1         0        1\n  0   0   0   0  0\n")
    content.append("SBOND\n  0 0 0 0\n")

    # Colors and Radii - Extended Defaults + Specifics from user example
    cpk_colors = {
        'H': (255, 255, 255), 'He': (217, 255, 255), 'Li': (204, 128, 255), 'Be': (194, 255, 0),
        'B': (255, 181, 181), 'C': (144, 144, 144), 'N': (48, 80, 248), 'O': (254, 3, 0),
        'F': (144, 224, 80), 'Ne': (179, 227, 245), 'Na': (171, 92, 242), 'Mg': (138, 255, 0),
        'Al': (191, 166, 166), 'Si': (240, 200, 160), 'P': (255, 128, 0), 'S': (255, 255, 48),
        'Cl': (31, 240, 31), 'Ar': (128, 209, 227), 'K': (143, 64, 212), 'Ca': (61, 255, 0),
        'Sc': (230, 230, 230), 'Ti': (120, 202, 254), 'V': (166, 166, 171), 'Cr': (138, 153, 199),
        'Mn': (156, 122, 199), 'Fe': (224, 102, 51), 'Co': (240, 144, 160), 'Ni': (80, 208, 80),
        'Cu': (200, 128, 51), 'Zn': (125, 128, 176), 'Ga': (194, 143, 143), 'Ge': (102, 143, 143),
        'As': (189, 128, 227), 'Se': (255, 161, 0), 'Br': (166, 41, 41), 'Kr': (92, 184, 209),
        'Rb': (112, 46, 176), 'Sr': (0, 255, 38), 'Y': (148, 255, 255), 'Zr': (148, 224, 224),
    }
    cpk_radii = {
         'H': 0.8, 'C': 1.0, 'N': 1.0, 'O': 0.74, 'F': 1.0, 'Si': 1.18, 'P': 1.10, 'S': 1.04,
         'Cl': 1.0, 'K': 2.35, 'Ca': 1.97, 'Ti': 1.47, 'Fe': 1.3, 'Co': 1.25, 'Ni': 1.24,
         'Cu': 1.28, 'Zn': 1.33, 'Ga': 1.22, 'Ge': 1.22, 'As': 1.2, 'Se': 1.17, 'Br': 1.14,
         'Sr': 2.15, 'Zr': 1.6, 'Mo': 1.4,
    }

    # SITET
    content.append("SITET\n")
    for i, atom in enumerate(atoms):
        el = atom['element']
        label = atom['label']
        rad = cpk_radii.get(el, 1.2)
        r, g, b_col = cpk_colors.get(el, (200, 200, 200))
        # CORRECTED uses 200 200 200 for the second color triplet in SITET
        content.append(f"  {i+1:<8} {label:<4} {rad:.4f} {r} {g} {b_col} 200 200 200 204  0\n")
    content.append("  0 0 0 0 0 0\n")

    # Vectors Placeholders (Matching VESTA structure)
    content.append("VECTR\n 0 0 0 0 0\n")
    content.append("VECTT\n 0 0 0 0 0\n")
    content.append("SPLAN\n  0   0   0   0\n")

    # Footer
    content.append("LBLAT\n -1\nLBLSP\n -1\nDLATM\n -1\nDLBND\n -1\nDLPLY\n -1\nPLN2D\n  0   0   0   0\nATOMT\n")
    
    # ATOMT
    for i, el in enumerate(unique_elements):
         rad = cpk_radii.get(el, 1.2)
         r, g, b_col = cpk_colors.get(el, (200, 200, 200))
         content.append(f"  {i+1:<9} {el:<3} {rad:.4f} {r} {g} {b_col} {r} {g} {b_col} 204\n")
    content.append("  0 0 0 0 0 0\n")

    # Static Footer Block
    footer_static = """SCENE
 1.000000  0.000000  0.000000  0.000000
 0.000000  1.000000  0.000000  0.000000
 0.000000  0.000000  1.000000  0.000000
 0.000000  0.000000  0.000000  1.000000
  0.000   0.000
  0.000
  1.000
HBOND 0 2

STYLE
DISPF 37753794
MODEL   0  1  0
SURFS   0  1  1
SECTS  32  1
FORMS   0  1
ATOMS   0  0  1
BONDS   1
POLYS   1
VECTS 1.000000
FORMP
  1  1.0   0   0   0
ATOMP
 24  24   0  50  2.0   0
BONDP
  1  16  0.250  2.000 127 127 127
POLYP
 204 1  1.000 180 180 180
ISURF
  0   0   0   0
TEX3P
  1  0.00000E+00  1.00000E+00
SECTP
  1  5.00000E-01  5.00000E-01  0.00000E+00  0.00000E+00  0.00000E+00  0.00000E+00
CONTR
 0.1 -1 1 1 10 -1 2 5
 2 1 2 1
   0   0   0
   0   0   0
   0   0   0
   0   0   0
HKLPP
 192 0  0.000 254 204 102
UCOLP
   0   1  1.000   0   0   0
COMPS 1
LABEL 1    12  1.000 0
PROJT 0  0.962
BKGRC
 255 255 255
DPTHQ 1 -0.5000  3.5000
LIGHT0 1
 1.000000  0.000000  0.000000  0.000000
 0.000000  1.000000  0.000000  0.000000
 0.000000  0.000000  1.000000  0.000000
 0.000000  0.000000  0.000000  1.000000
 0.000000  0.000000 20.000000  0.000000
 0.000000  0.000000 -1.000000
  26  26  26 255
 179 179 179 255
 255 255 255 255
LIGHT1
 1.000000  0.000000  0.000000  0.000000
 0.000000  1.000000  0.000000  0.000000
 0.000000  0.000000  1.000000  0.000000
 0.000000  0.000000  0.000000  1.000000
 0.000000  0.000000 20.000000  0.000000
 0.000000  0.000000 -1.000000
   0   0   0   0
   0   0   0   0
   0   0   0   0
LIGHT2
 1.000000  0.000000  0.000000  0.000000
 0.000000  1.000000  0.000000  0.000000
 0.000000  0.000000  1.000000  0.000000
 0.000000  0.000000  0.000000  1.000000
 0.000000  0.000000 20.000000  0.000000
 0.000000  0.000000 -1.000000
   0   0   0   0
   0   0   0   0
   0   0   0   0
LIGHT3
 1.000000  0.000000  0.000000  0.000000
 0.000000  1.000000  0.000000  0.000000
 0.000000  0.000000  1.000000  0.000000
 0.000000  0.000000  0.000000  1.000000
 0.000000  0.000000 20.000000  0.000000
 0.000000  0.000000 -1.000000
   0   0   0   0
   0   0   0   0
   0   0   0   0
SECCL 0

TEXCL 0

ATOMM
   0   0   0 255
  25.600
BONDM
   0   0   0 255
 128.000
POLYM
 255 255 255 255
 128.000
SURFM
   0   0   0 255
 128.000
FORMM
 255 255 255 255
 128.000
HKLPM
 255 255 255 255
 128.000
"""
    content.append(footer_static)

    return "".join(content)

def _generate_mcif_content(title, lattice, atoms, vectors):
    """
    Generates the content for a magnetic CIF (.mcif) file.
    """
    a, b, c, alpha, beta, gamma = _lattice_params(lattice)
    
    content = [
        "#======================================================================",
        "# CRYSTAL DATA (generated from VESTA style structure)",
        "#----------------------------------------------------------------------",
        "data_VESTA_phase_1",
        "",
        f"_chemical_name_common                  '{title}'",
        f"_cell_length_a                         {a:.6f}",
        f"_cell_length_b                         {b:.6f}",
        f"_cell_length_c                         {c:.6f}",
        f"_cell_angle_alpha                      {alpha:.6f}",
        f"_cell_angle_beta                       {beta:.6f}",
        f"_cell_angle_gamma                      {gamma:.6f}",
        "_space_group_name_H-M_alt              'P 1'",
        "_space_group_IT_number                 1",
        "",
        "loop_",
        "_space_group_symop_operation_xyz",
        "   'x, y, z'",
        "",
        "loop_",
        "   _atom_site_label",
        "   _atom_site_occupancy",
        "   _atom_site_fract_x",
        "   _atom_site_fract_y",
        "   _atom_site_fract_z",
        "   _atom_site_adp_type",
        "   _atom_site_U_iso_or_equiv",
        "   _atom_site_type_symbol",
    ]
    
    for atom in atoms:
        lbl = atom['label']
        el = atom['element']
        x, y, z = atom['coords']
        # Format: label occ x y z adp iso type
        content.append(f"   {lbl:<10} 1.000 {x:12.6f} {y:12.6f} {z:12.6f}    Uiso  ? {el}")
        
    content.append("")
    content.append("loop_")
    content.append("   _atom_site_moment_label")
    content.append("   _atom_site_moment_crystalaxis_x")
    content.append("   _atom_site_moment_crystalaxis_y")
    content.append("   _atom_site_moment_crystalaxis_z")
    
    for atom, vec in zip(atoms, vectors):
        lbl = atom['label']
        mx, my, mz = vec
        content.append(f"   {lbl:<10} {mx:12.6f} {my:12.6f} {mz:12.6f}")

    content.append("")
    return "\n".join(content)
#</editor-fold>

def _parse_yaml_for_arrows(band_yaml_path: Path):
    """Parses the band.yaml file using pyyaml."""
    if not band_yaml_path.exists():
        raise FileNotFoundError(f"Phonopy band file '{band_yaml_path}' not found.")
    with open(band_yaml_path, 'r') as f:
        data = yaml.safe_load(f)
    if 'phonon' not in data:
        raise ValueError(f"'{band_yaml_path}' does not appear to be a valid Phonopy band.yaml file.")
    return data

def _get_special_points_map(phonopy_data):
    """Creates a map of q-point index to special point label."""
    if 'labels' not in phonopy_data or 'segment_nqpoint' not in phonopy_data:
        return {}
        
    segments = phonopy_data['segment_nqpoint']
    labels = phonopy_data['labels']
    
    # print(f"[ARROW_DEBUG] Found {len(segments)} segments and {len(labels)} label pairs.")

    special_points = {}
    current_idx = 0
    
    for i, seg_len in enumerate(segments):
        if i < len(labels):
            start_label, end_label = [l.upper().replace("GAMMA", "GM") for l in labels[i]]
            special_points[current_idx] = start_label
            end_idx = current_idx + seg_len - 1
            special_points[end_idx] = end_label
            
        current_idx += seg_len
        
    # print(f"[ARROW_DEBUG] Generated special_points map: {special_points}")
    return special_points

def _parse_irreps_yaml(irreps_yaml_path: Path):
    """Parses irreps.yaml to map band indices to labels."""
    if not irreps_yaml_path or not irreps_yaml_path.exists():
        return None, {}
    
    try:
        with open(irreps_yaml_path, 'r') as f:
            data = yaml.safe_load(f)
        
        q_pos = data.get("q-position")
        normal_modes = data.get("normal_modes", [])
        label_map = {}
        for mode in normal_modes:
            label = mode.get("ir_label")
            indices = mode.get("band_indices", [])
            if label and indices:
                for idx in indices:
                    label_map[idx] = label
        return q_pos, label_map
    except Exception as e:
        print(f"Warning: Failed to parse irreps file: {e}", file=sys.stderr)
        return None, {}

def write_vesta_files_for_arrows(
    band_yaml_path: Path,
    poscar_path: Path,
    output_dir: Path,
    arrow_length: float = 1.5,
    arrow_min_cutoff: float = 0.1,
    arrow_qpoint_gamma: bool = False,
    qpoint_index: int | None = None,
    irreps_yaml_path: Path | None = None,
    target_q_point: list[float] | None = None,
):
    """
    Extracts phonon modes from yaml and writes VESTA files for visualization.
    Generates a base .vesta file from the POSCAR first, then uses it as a template.
    """
    try:
        title, lattice, atoms, unique_elements = _read_poscar_for_vesta(poscar_path)
        base_vesta_content = _generate_vesta_content(title, lattice, atoms, unique_elements)
        
        base_vesta_path = output_dir / f"{poscar_path.stem}.vesta"
        with open(base_vesta_path, 'w') as f:
            f.write(base_vesta_content)
        print(f"  Generated base VESTA file: {base_vesta_path}")

        phonopy_data = _parse_yaml_for_arrows(band_yaml_path)
        irreps_q_pos, irreps_label_map = _parse_irreps_yaml(irreps_yaml_path)

    except (FileNotFoundError, ValueError) as e:
        print(f"Error: {e}", file=sys.stderr)
        return

    phonons = phonopy_data.get('phonon', [])
    if not phonons:
        print("Warning: No phonon data found in band.yaml.", file=sys.stderr)
        return

    # --- Determine Mode and Filter Q-points ---
    special_points_map = {i: lbl for i, lbl in enumerate(phonopy_data.get('labels', [])) if lbl}
    
    indices_to_process, labels_for_q_indices = [], {}
    if target_q_point is not None:
        print(f"  Mode: Writing arrows for specific target q-point: {target_q_point}")
        # Pinpoint calculation (usually from qpoints.yaml)
        if len(phonons) == 1:
            indices_to_process = [0]
            if 0 in special_points_map:
                labels_for_q_indices[0] = special_points_map[0]
        else:
            for idx, ph in enumerate(phonons):
                q_pos = ph.get('q-position')
                if q_pos and np.allclose(q_pos, target_q_point, atol=1e-4):
                    if idx not in indices_to_process:
                        indices_to_process.append(idx)
                        if idx in special_points_map:
                            labels_for_q_indices[idx] = special_points_map[idx]
    elif qpoint_index is not None:
        print(f"  Mode: Writing arrows for specified q-point index: {qpoint_index}")
        if 0 <= qpoint_index < len(phonons):
            indices_to_process.append(qpoint_index)
        else:
            print(f"Error: q-point index {qpoint_index} is out of range (0-{len(phonons)-1}).")
            return
    elif arrow_qpoint_gamma:
        print("  Mode: Writing arrows for Gamma point only.")
        for idx, ph in enumerate(phonons):
            if np.linalg.norm(ph.get('q-position', [1,1,1])) < 1e-4:
                if idx not in indices_to_process:
                    indices_to_process.append(idx)
                    labels_for_q_indices[idx] = "GAMMA"
    else: 
        if special_points_map:
            print("  Mode: Writing arrows for special q-points (default).")
            indices_to_process = sorted(special_points_map.keys())
            labels_for_q_indices = special_points_map
        else:
             print("  Mode: Writing arrows for all q-points in file (default).")
             indices_to_process = list(range(len(phonons)))

    if not indices_to_process:
        print("Warning: No q-points matched the specified criteria.")
        return
    
    print(f"  Preparing to write VESTA files for {len(indices_to_process)} q-points...")
    THZ_TO_CM1, natoms = 33.3564095, phonopy_data.get('natom', 0)
    
    with open(base_vesta_path, 'r') as f:
        template_content = f.read()

    for i, q_idx in enumerate(indices_to_process):
        phonon_qpoint = phonons[q_idx]
        q_pos = phonon_qpoint.get('q-position')
        q_pos_str = '%.2f_%.2f_%.2f' % (q_pos[0], q_pos[1], q_pos[2])
        
        folder_idx = i + 1
        
        # When processing a specific target q-point, write directly to the output_dir.
        # Otherwise, create sub-directories for each q-point from the band path.
        current_output_dir = output_dir
        if target_q_point is None and len(indices_to_process) > 1:
            label = labels_for_q_indices.get(q_idx, "QP")
            current_output_dir = output_dir / f"QPOINTS_{folder_idx:03d}-{label}={q_pos_str}"
        
        current_output_dir.mkdir(exist_ok=True, parents=True)
        
        is_irreps_qpoint = False
        if irreps_q_pos is not None and np.allclose(q_pos, irreps_q_pos, atol=1e-4):
            is_irreps_qpoint = True

        for band_idx, band_data in enumerate(phonon_qpoint.get('band', [])):
            eigenvectors = band_data.get('eigenvector')
            if not eigenvectors:
                continue

            # New scaling logic: per-mode normalization
            displacements = np.array([[v[0] for v in eig_vec] for eig_vec in eigenvectors])
            lengths = np.linalg.norm(displacements, axis=1)
            max_disp_length = np.max(lengths) if lengths.size > 0 else 0
            
            scale_factor = 1.0
            if max_disp_length > 1e-8: # Avoid division by zero for non-moving modes
                scale_factor = arrow_length / max_disp_length

            vectr_block = "VECTR\n"
            scaled_vectors = []
            for atom_idx, disp_vec in enumerate(displacements):
                scaled_disp = disp_vec * scale_factor
                if np.linalg.norm(scaled_disp) < arrow_min_cutoff:
                    scaled_disp = np.zeros(3)
                
                scaled_vectors.append(scaled_disp)
                vectr_block += f"{atom_idx+1:>5d}{scaled_disp[0]:10.5f}{scaled_disp[1]:10.5f}{scaled_disp[2]:10.5f}\n"
                vectr_block += f"{atom_idx+1:>5d} 0 0 0 0\n  0 0 0 0 0\n"
            vectr_block += '0 0 0 0 0\n'

            freq_thz = band_data.get('frequency', 0)
            freq_cm1 = freq_thz * THZ_TO_CM1
            filename_q_label = labels_for_q_indices.get(q_idx, q_pos_str)
            
            ir_label_part = ""
            if is_irreps_qpoint:
                ir_label = irreps_label_map.get(band_idx + 1)
                if ir_label:
                    ir_label_part = f"-{ir_label}"
            
            filename_base = f"MODE{band_idx+1:03d}-QPOINTS={filename_q_label}{ir_label_part}-freq={freq_cm1:.2f}cm-1"
            vesta_filename = f"{filename_base}.vesta"
            mcif_filename = f"{filename_base}.mcif"
            
            vectt_block = "VECTT\n"
            for atom_idx in range(natoms):
                vectt_block += f"{atom_idx+1:>5d}  0.500 255   0   0 0\n"
            vectt_block += '0 0 0 0 0\n'
            
            parts = template_content.split('VECTR', 1)
            pre_vectr = parts[0]
            subparts = parts[1].split('SPLAN', 1)
            post_splan = subparts[1]
            
            new_content = pre_vectr + vectr_block + vectt_block + "SPLAN" + post_splan
            
            with open(current_output_dir / vesta_filename, 'w') as f:
                f.write(new_content)
                
            # Generate and write MCIF
            mcif_content = _generate_mcif_content(title, lattice, atoms, scaled_vectors)
            with open(current_output_dir / mcif_filename, 'w') as f:
                f.write(mcif_content)

    print(f"  Done. VESTA and MCIF files for arrows are written in {output_dir}")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
seekpath2bandconf.py
Make phonopy band.conf directly from POSCAR using SeeK-path (no VASPKIT needed).

- Robust POSCAR parser (VASP4/5, Selective dynamics, Direct/Cartesian)
- POSCAR -> SeeK-path standard k-path (continuous chain)
- DIM: from --dim "a b c" (highest priority) or phonopy_disp.yaml if present; otherwise blank
- BAND: single line, one k-point per node (junctions deduped)
- BAND_LABELS: cleaned (GAMMA->GM by default, underscores removed)
- ATOM_NAME override: --atom-names "K Zr P O" or --rename "Na=K"
- Summary print: POSCAR used, space group, q-path & coords, DIM source

Usage
  pip install seekpath spglib
  python seekpath2bandconf.py --poscar POSCAR --yaml phonopy_disp.yaml --out band.conf
  # optional:
  #   --atom-names "K Zr P O"
  #   --rename "Na=K"
  #   --gamma "Γ"
  #   --symprec 1e-5
  #   --dim "3 3 3"
  #   --no-defaults
"""

import argparse
import re
from pathlib import Path

# (optional) silence DeprecationWarnings from spglib
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


# ------------------------- POSCAR I/O -------------------------

def read_poscar(poscar_path: Path):
    with poscar_path.open("r", encoding="utf-8") as f:
        raw = [ln.rstrip("\n") for ln in f]

    def next_nonempty(i):
        while i < len(raw) and raw[i].strip() == "":
            i += 1
        return i

    i = next_nonempty(0)
    if i >= len(raw):
        raise ValueError("POSCAR is empty")

    comment = raw[i]; i = next_nonempty(i+1)
    if i >= len(raw):
        raise ValueError("POSCAR: missing scale line")
    scale = float(raw[i].split()[0]); i = next_nonempty(i+1)

    # lattice vectors
    lat = []
    for _ in range(3):
        if i >= len(raw):
            raise ValueError("POSCAR: missing lattice vectors")
        parts = raw[i].split()
        if len(parts) < 3:
            raise ValueError("POSCAR: lattice vector line has < 3 numbers")
        vec = [float(x) for x in parts[:3]]
        lat.append([scale * v for v in vec])
        i = next_nonempty(i+1)

    if i >= len(raw):
        raise ValueError("POSCAR: missing symbols/counts line")

    # Detect VASP5 (symbols line) vs VASP4
    tokens = raw[i].split()

    def is_number(x):
        try:
            float(x); return True
        except Exception:
            return False

    vasp5 = (tokens and any(not is_number(t) for t in tokens))
    if vasp5:
        symbols = tokens[:]                   # e.g. Na Zr P O
        i = next_nonempty(i+1)
        if i >= len(raw):
            raise ValueError("POSCAR: missing counts line after symbols")
        counts = [int(x) for x in raw[i].split()]
        i = next_nonempty(i+1)
        if len(symbols) != len(counts):
            if len(symbols) > len(counts):
                symbols = symbols[:len(counts)]
            else:
                symbols = symbols + [f"E{j+1}" for j in range(len(counts)-len(symbols))]
    else:
        # VASP4: counts here, no symbols line
        counts = [int(x) for x in tokens]
        symbols = []  # no symbols; ATOM_NAME will be blank unless overridden
        i = next_nonempty(i+1)

    # Optional "Selective dynamics"
    if i < len(raw) and raw[i].strip().lower().startswith("selective"):
        i = next_nonempty(i+1)

    # Coordinate type
    if i >= len(raw):
        raise ValueError("POSCAR: missing coordinate type line")
    ctok = raw[i].strip().lower()
    direct = ctok.startswith("d")
    cart = ctok.startswith("c")
    if not (direct or cart):
        raise ValueError(f"POSCAR: unknown coordinate type line: {raw[i]}")
    i = next_nonempty(i+1)

    # Atom coordinates
    nat = sum(counts)
    coord_lines, read = [], 0
    while i < len(raw) and read < nat:
        if raw[i].strip() != "":
            coord_lines.append(raw[i])
            read += 1
        i += 1
    if read < nat:
        raise ValueError("POSCAR: not enough atomic coordinate lines")

    if direct:
        frac = []
        for ln in coord_lines:
            parts = ln.split()
            x, y, z = [float(u) for u in parts[:3]]
            frac.append([x, y, z])
    else:
        import numpy as np
        A = np.array(lat).T  # columns a,b,c
        Ainv = np.linalg.inv(A)
        frac = []
        for ln in coord_lines:
            parts = ln.split()
            cx, cy, cz = [float(u) for u in parts[:3]]
            f = Ainv @ (scale * np.array([cx, cy, cz]))
            frac.append(f.tolist())

    # kinds: build from counts
    kinds = []
    nsp = len(counts)
    for si in range(nsp):
        kinds.extend([si + 1] * counts[si])

    return {
        "comment": comment,
        "lattice": lat,
        "frac": frac,
        "kinds": kinds,
        "symbols": symbols,  # may be []
        "counts": counts,
    }


# ------------------------- phonopy_disp.yaml -> DIM -------------------------

def read_dim_from_yaml(yaml_path: Path):
    """Return [a,b,c] or None if not found or file missing."""
    try:
        return _read_dim_yaml_inner(yaml_path)
    except FileNotFoundError:
        return None

def _read_dim_yaml_inner(yaml_path: Path):
    dim = None
    pat_dim = re.compile(r'^\s*dim:\s*"([^"]+)"\s*$', re.IGNORECASE)
    pat_row = re.compile(r'^\s*-\s*\[\s*(-?\d+)\s*,\s*(-?\d+)\s*,\s*(-?\d+)\s*\]\s*$')

    with yaml_path.open("r", encoding="utf-8") as f:
        for ln in f:
            m = pat_dim.match(ln)
            if m:
                parts = m.group(1).split()
                if len(parts) == 3 and all(p.lstrip("-").isdigit() for p in parts):
                    dim = [int(p) for p in parts]
                break
    if dim is not None:
        return dim

    rows, in_super = [], False
    with yaml_path.open("r", encoding="utf-8") as f:
        for ln in f:
            if "supercell_matrix" in ln:
                in_super = True
                continue
            if in_super:
                if ln.strip().startswith("- ["):
                    m = pat_row.match(ln)
                    if m:
                        rows.append([int(m.group(i)) for i in range(1, 4)])
                else:
                    break

    if rows:
        if all(rows[i][j] == (rows[i][i] if i == j else 0) for i in range(3) for j in range(3)):
            dim = [abs(rows[i][i]) for i in range(3)]
        else:
            import math
            dim = [max(1, int(round(math.sqrt(sum(c*c for c in r))))) for r in rows[:3]]
            print("[WARN] Non-diagonal supercell_matrix; approximated DIM =", dim)

    return dim  # may be None


# ------------------------- helpers -------------------------

def _fmt(x):
    v = float(x)
    if abs(v) < 1e-12:
        v = 0.0
    s = f"{v:.3f}"
    if s == "-0.000":
        s = "0.000"
    return s

def _clean_label(lbl: str, gamma="GM"):
    if lbl.upper() == "GAMMA":
        return gamma
    return lbl.replace("_", "")


# ------------------------- label chain & band -------------------------

def build_label_chain(path_segments):
    if not path_segments:
        return []
    chain = []
    s0, e0 = path_segments[0]
    chain.append(s0); chain.append(e0)
    for (s, e) in path_segments[1:]:
        if chain[-1] != s:
            chain.append(s)
        chain.append(e)
    dedup = [chain[0]]
    for lab in chain[1:]:
        if lab != dedup[-1]:
            dedup.append(lab)
    return dedup

def band_points_one_line_from_seekpath(path_data, gamma_label="GM"):
    pc = path_data["point_coords"]
    segs = path_data["path"]
    chain = build_label_chain(segs)
    labels = [_clean_label(x, gamma_label) for x in chain]
    pts = [f"{_fmt(pc[lab][0])} {_fmt(pc[lab][1])} {_fmt(pc[lab][2])}" for lab in chain]
    band_line = "BAND = " + "    ".join(pts)
    return band_line, labels, chain  # chain: raw labels for summary


# ------------------------- atom-name override -------------------------

def parse_atom_override(args, symbols_from_poscar):
    if args.atom_names:
        return args.atom_names.split()
    if args.rename:
        ren = {}
        for pair in args.rename.split(","):
            old, new = pair.split("=")
            ren[old.strip()] = new.strip()
        if symbols_from_poscar:
            return [ren.get(s, s) for s in symbols_from_poscar]
        else:
            return []
    return symbols_from_poscar


# ------------------------- pretty summary -------------------------

def print_summary(poscar_path: Path, path_data, chain_labels, gamma_cleaned_labels, dim, dim_source):
    try:
        poscar_disp = str(poscar_path.resolve())
    except Exception:
        poscar_disp = str(poscar_path)

    sg_int = path_data.get("spacegroup_international", "?")
    sg_no  = path_data.get("spacegroup_number", "?")
    bravais = path_data.get("bravais_lattice", "?")

    print("------------------------------------------------------------")
    print(f"[Seekpath] POSCAR: {poscar_disp}")
    print(f"[Seekpath] Space group: {sg_int} (No.{sg_no}), Bravais: {bravais}")
    print(f"[Seekpath] Q-path (labels): {' - '.join(gamma_cleaned_labels)}")
    print("[Seekpath] Q-points (reciprocal crystal units):")
    pc = path_data["point_coords"]
    for raw_lab, clean_lab in zip(chain_labels, gamma_cleaned_labels):
        k = pc[raw_lab]
        print(f"  {clean_lab:>4s} : {_fmt(k[0])}  {_fmt(k[1])}  {_fmt(k[2])}")
    print(f"[Seekpath] Total q-points: {len(chain_labels)}")

    if dim is None:
        print("[Info] DIM not set (no phonopy_disp.yaml). The output band.conf will contain 'DIM =' (blank).")
        print("       Provide --yaml phonopy_disp.yaml or use --dim \"a b c\" to set it explicitly.")
    else:
        print(f"[Seekpath] DIM = {dim[0]} {dim[1]} {dim[2]}  (source: {dim_source})")
    print("------------------------------------------------------------")


# ------------------------- main -------------------------

def main():
    ap = argparse.ArgumentParser(description="Make phonopy band.conf via SeeK-path (one-line BAND)")
    ap.add_argument("--poscar", default="POSCAR", type=Path)
    ap.add_argument("--yaml", default="phonopy_disp.yaml", type=Path)
    ap.add_argument("--out", default="band.conf", type=Path)
    ap.add_argument("--gamma", default="GM", help="Gamma label for BAND_LABELS (e.g., GM or Γ)")
    ap.add_argument("--symprec", type=float, default=1e-5, help="Symmetry tolerance passed to SeeK-path (default: 1e-5)")
    ap.add_argument("--dim", default=None, help='Override DIM as a string "a b c" (e.g., "3 3 3")')
    ap.add_argument("--no-defaults", action="store_true")
    ap.add_argument("--atom-names", default=None, help='Override ATOM_NAME, e.g. "K Zr P O"')
    ap.add_argument("--rename", default=None, help='Rename mapping, e.g. "Na=K,Zr=Zr"')
    args = ap.parse_args()

    # POSCAR -> cell
    pos = read_poscar(args.poscar)
    atom_names = parse_atom_override(args, pos["symbols"])
    lattice, positions, numbers = pos["lattice"], pos["frac"], pos["kinds"]
    cell = (lattice, positions, numbers)

    # SeeK-path (with symprec)
    import seekpath
    path_data = seekpath.get_path(cell, symprec=args.symprec)

    band_line, labels, chain_raw = band_points_one_line_from_seekpath(path_data, gamma_label=args.gamma)

    # DIM priority: --dim > --yaml (if exists) > None (blank)
    dim = None
    dim_source = None
    if args.dim:
        parts = args.dim.split()
        if len(parts) == 3 and all(p.lstrip("-").isdigit() for p in parts):
            dim = [int(p) for p in parts]
            dim_source = "--dim"
        else:
            print("[WARN] --dim must be like: --dim \"3 3 3\"; ignoring it.")
    if dim is None:
        if args.yaml and Path(args.yaml).exists():
            dim = read_dim_from_yaml(Path(args.yaml))
            dim_source = "phonopy_disp.yaml" if dim is not None else None
        else:
            dim = None

    # Write band.conf
    out_lines = []
    if atom_names:
        out_lines.append(f"ATOM_NAME = {' '.join(atom_names)}")
    else:
        out_lines.append("ATOM_NAME =")
    if dim is None:
        out_lines.append("DIM =")  # leave blank as requested
    else:
        out_lines.append(f"DIM = {dim[0]} {dim[1]} {dim[2]}")
    out_lines.append(band_line)
    out_lines.append(f"BAND_LABELS = {' '.join(labels)}")
    if not args.no_defaults:
        out_lines.append("FORCE_SETS = READ")
        out_lines.append("FC_SYMMETRY = .TRUE.")
        out_lines.append("#EIGENVECTORS = .TRUE.")

    Path(args.out).write_text("\n".join(out_lines) + "\n", encoding="utf-8")
    print(f"[OK] Wrote {args.out}")

    # Summary
    print_summary(args.poscar, path_data, chain_raw, labels, dim, dim_source)


if __name__ == "__main__":
    main()




from typing import List, NamedTuple, Optional, Tuple, Union

from opt_einsum_fx import optimize_einsums_full
import torch
from torch import fx

import macer.externals.e3nn_legacy as e3nn
from macer.externals.e3nn_legacy import o3
from macer.externals.e3nn_legacy.util import prod
from macer.externals.e3nn_legacy.util.codegen import CodeGenMixin
from macer.externals.e3nn_legacy.util.jit import compile_mode

from ._tensor_product._codegen import _sum_tensors


class Instruction(NamedTuple):
    i_in: int
    i_out: int
    path_shape: tuple
    path_weight: float


@compile_mode('script')
class Linear(CodeGenMixin, torch.nn.Module):
    r"""Linear operation equivariant to :math:`O(3)`

    Notes
    -----
        `e3nn.o3.Linear` objects created with different partitionings of the same irreps, such as ``Linear("10x0e", "0e")`` and ``Linear("3x0e + 7x0e", "0e")``, are *not* equivalent: the second module has more instructions, which affects normalization. In a rough sense:

            Linear("10x0e", "0e") = normalization_coeff_0 * W_0 @ input
            Linear("3x0e + 7x0e", "0e") = normalization_coeff_1 * W_1 @ input[:3] + normalization_coeff_2 * W_2 @ input[3:]

        To make them equivalent, simplify ``irreps_in`` before constructing network modules:

            o3.Irreps("3x0e + 7x0e").simplify()  # => 10x0e


    Parameters
    ----------
    irreps_in : `e3nn.o3.Irreps`
        representation of the input

    irreps_out : `e3nn.o3.Irreps`
        representation of the output

    internal_weights : bool
        whether the `e3nn.o3.Linear` should store its own weights. Defaults to ``True`` unless ``shared_weights`` is explicitly set to ``False``, for consistancy with `e3nn.o3.TensorProduct`.

    shared_weights : bool
        whether the `e3nn.o3.Linear` should be weighted individually for each input in a batch. Defaults to ``False``. Cannot be ``True`` if ``internal_weights`` is ``True``.

    instructions : list of 2-tuples, optional
        list of tuples ``(i_in, i_out)`` indicating which irreps in ``irreps_in`` should contribute to which irreps in ``irreps_out``. If ``None`` (the default), all allowable instructions will be created: every ``(i_in, i_out)`` such that ``irreps_in[i_in].ir == irreps_out[i_out].ir``.

    biases : list of bool, optional
        indicates for each element of ``irreps_out`` if it has a bias. By default there is no bias. If ``biases=True`` it gives bias to all scalars (l=0 and p=1).

    Attributes
    ----------
    weight_numel : int
        the size of the weights for this `e3nn.o3.Linear`

    Examples
    --------
    Linearly combines 4 scalars into 8 scalars and 16 vectors into 8 vectors.

    >>> lin = Linear("4x0e+16x1o", "8x0e+8x1o")
    >>> lin.weight_numel
    160

    Create a "block sparse" linear that does not combine two different groups of scalars;
    note that the number of weights is 4*4 + 3*3 = 25:

    >>> lin = Linear("4x0e + 3x0e", "4x0e + 3x0e", instructions=[(0, 0), (1, 1)])
    >>> lin.weight_numel
    25

    Be careful: because they have different instructions, the following two operations are not normalized in the same way, even though they contain all the same "connections":

    >>> lin1 = Linear("10x0e", "0e")
    >>> lin2 = Linear("3x0e + 7x0e", "0e")
    >>> lin1.weight_numel == lin2.weight_numel
    True
    >>> with torch.no_grad():
    ...     lin1.weight.fill_(1.0)
    ...     lin2.weight.fill_(1.0)
    Parameter containing:
    ...
    >>> x = torch.arange(10.0)
    >>> (lin1(x) - lin2(x)).abs().item() < 1e-5
    True

    """
    weight_numel: int
    internal_weights: bool
    shared_weights: bool

    def __init__(
        self,
        irreps_in,
        irreps_out,
        *,
        f_in: Optional[int] = None,
        f_out: Optional[int] = None,
        internal_weights: Optional[bool] = None,
        shared_weights: Optional[bool] = None,
        instructions: Optional[List[Tuple[int, int]]] = None,
        biases: Union[bool, List[bool]] = False,
        path_normalization: str = 'element',
        _optimize_einsums: Optional[bool] = None
    ):
        super().__init__()

        assert path_normalization in ['element', 'path']

        irreps_in = o3.Irreps(irreps_in)
        irreps_out = o3.Irreps(irreps_out)

        if instructions is None:
            # By default, make all possible connections
            instructions = [
                (i_in, i_out)
                for i_in, (_, ir_in) in enumerate(irreps_in)
                for i_out, (_, ir_out) in enumerate(irreps_out)
                if ir_in == ir_out
            ]

        instructions = [
            Instruction(
                i_in=i_in,
                i_out=i_out,
                path_shape=(irreps_in[i_in].mul, irreps_out[i_out].mul),
                path_weight=1,
            )
            for i_in, i_out in instructions
        ]

        def alpha(ins):
            x = sum(
                irreps_in[i.i_in if path_normalization == 'element' else ins.i_in].mul
                for i in instructions
                if i.i_out == ins.i_out
            )
            if f_in is not None:
                x *= f_in
            return 1.0 if x == 0 else x

        instructions = [
            Instruction(
                i_in=ins.i_in,
                i_out=ins.i_out,
                path_shape=ins.path_shape,
                path_weight=alpha(ins)**(-0.5)
            )
            for ins in instructions
        ]

        for ins in instructions:
            if not ins.i_in < len(irreps_in):
                raise IndexError(f"{ins.i_in} is not a valid index for irreps_in")
            if not ins.i_out < len(irreps_out):
                raise IndexError(f"{ins.i_out} is not a valid index for irreps_out")
            if not (ins.i_in == -1 or irreps_in[ins.i_in].ir == irreps_out[ins.i_out].ir):
                raise ValueError(
                    f"{ins.i_in} and {ins.i_out} do not have the same irrep"
                )

        if biases is None:
            biases = len(irreps_out) * (False,)
        if isinstance(biases, bool):
            biases = [biases and ir.is_scalar() for _, ir in irreps_out]

        assert len(biases) == len(irreps_out)
        assert all(ir.is_scalar() or (not b) for b, (_, ir) in zip(biases, irreps_out))

        instructions += [
            Instruction(
                i_in=-1,
                i_out=i_out,
                path_shape=(mul_ir.dim,),
                path_weight=1.0
            )
            for i_out, (bias, mul_ir) in enumerate(zip(biases, irreps_out)) if bias
        ]

        # == Process arguments ==
        if shared_weights is False and internal_weights is None:
            internal_weights = False

        if shared_weights is None:
            shared_weights = True

        if internal_weights is None:
            internal_weights = True

        assert shared_weights or not internal_weights
        self.internal_weights = internal_weights
        self.shared_weights = shared_weights

        self.irreps_in = irreps_in
        self.irreps_out = irreps_out
        self.instructions = instructions

        opt_defaults = e3nn.get_optimization_defaults()
        self._optimize_einsums = _optimize_einsums if _optimize_einsums is not None else opt_defaults['optimize_einsums']
        del opt_defaults

        # == Generate code ==
        graphmod, self.weight_numel, self.bias_numel = _codegen_linear(
            self.irreps_in,
            self.irreps_out,
            self.instructions,
            f_in,
            f_out,
            shared_weights=shared_weights,
            optimize_einsums=self._optimize_einsums
        )
        self._codegen_register({
            "_compiled_main": graphmod
        })

        # == Generate weights ==
        if internal_weights and self.weight_numel > 0:
            assert self.shared_weights, "Having internal weights impose shared weights"
            self.weight = torch.nn.Parameter(torch.randn(*((f_in, f_out) if f_in is not None else ()), self.weight_numel))
        else:
            # For TorchScript, there always has to be some kind of defined .weight
            self.register_buffer('weight', torch.Tensor())

        # == Generate biases ==
        if internal_weights and self.bias_numel > 0:
            assert self.shared_weights, "Having internal weights impose shared weights"
            self.bias = torch.nn.Parameter(torch.zeros(*((f_out,) if f_out is not None else ()), self.bias_numel))  # see appendix C.1 and Eq.5 of https://arxiv.org/pdf/2011.14522.pdf
        else:
            self.register_buffer('bias', torch.Tensor())

        # == Compute output mask ==
        if self.irreps_out.dim > 0:
            output_mask = torch.cat([
                torch.ones(mul_ir.dim)
                if any(
                    (ins.i_out == i_out) and (0 not in ins.path_shape)
                    for ins in self.instructions
                )
                else torch.zeros(mul_ir.dim)
                for i_out, mul_ir in enumerate(self.irreps_out)
            ])
        else:
            output_mask = torch.ones(0)
        self.register_buffer('output_mask', output_mask)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.irreps_in} -> {self.irreps_out} | {self.weight_numel} weights)"

    def forward(self, features, weight: Optional[torch.Tensor] = None, bias: Optional[torch.Tensor] = None):
        """evaluate

        Parameters
        ----------
        features : `torch.Tensor`
            tensor of shape ``(..., irreps_in.dim)``

        weight : `torch.Tensor`, optional
            required if ``internal_weights`` is `False`

        Returns
        -------
        `torch.Tensor`
            tensor of shape ``(..., irreps_out.dim)``
        """
        if weight is None:
            if self.weight_numel > 0 and not self.internal_weights:
                raise RuntimeError("Weights must be provided when internal_weights = False")
            weight = self.weight
        if bias is None:
            if self.bias_numel > 0 and not self.internal_weights:
                raise RuntimeError("Biases must be provided when internal_weights = False")
            bias = self.bias
        return self._compiled_main(features, weight, bias)

    def weight_view_for_instruction(
        self,
        instruction: int,
        weight: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        r"""View of weights corresponding to ``instruction``.

        Parameters
        ----------
        instruction : int
            The index of the instruction to get a view on the weights for.

        weight : `torch.Tensor`, optional
            like ``weight`` argument to ``forward()``

        Returns
        -------
        `torch.Tensor`
            A view on ``weight`` or this object's internal weights for the weights corresponding to the ``instruction`` th instruction.
        """
        if weight is None:
            assert self.internal_weights, "Weights must be provided when internal_weights = False"
            weight = self.weight
        batchshape = weight.shape[:-1]
        offset = sum(prod(ins.path_shape) for ins in self.instructions[:instruction])
        ins = self.instructions[instruction]
        return weight.narrow(-1, offset, prod(ins.path_shape)).view(batchshape + ins.path_shape)

    def weight_views(
        self,
        weight: Optional[torch.Tensor] = None,
        yield_instruction: bool = False
    ):
        r"""Iterator over weight views for all instructions.

        Parameters
        ----------
        weight : `torch.Tensor`, optional
            like ``weight`` argument to ``forward()``

        yield_instruction : `bool`, default False
            Whether to also yield the corresponding instruction.

        Yields
        ------
        If ``yield_instruction`` is ``True``, yields ``(instruction_index, instruction, weight_view)``.
        Otherwise, yields ``weight_view``.
        """
        if weight is None:
            assert self.internal_weights, "Weights must be provided when internal_weights = False"
            weight = self.weight
        batchshape = weight.shape[:-1]
        offset = 0
        for ins_i, ins in enumerate(self.instructions):
            flatsize = prod(ins.path_shape)
            this_weight = weight.narrow(-1, offset, flatsize).view(batchshape + ins.path_shape)
            offset += flatsize
            if yield_instruction:
                yield ins_i, ins, this_weight
            else:
                yield this_weight


def _codegen_linear(
    irreps_in: o3.Irreps,
    irreps_out: o3.Irreps,
    instructions: List[Instruction],
    f_in: Optional[int] = None,
    f_out: Optional[int] = None,
    shared_weights: bool = False,
    optimize_einsums: bool = True,
) -> Tuple[fx.GraphModule, int, int]:
    graph_out = fx.Graph()
    tracer_out = fx.proxy.GraphAppendingTracer(graph_out)

    # = Function definitions =
    x = fx.Proxy(graph_out.placeholder('x', torch.Tensor), tracer_out)
    ws = fx.Proxy(graph_out.placeholder('w', torch.Tensor), tracer_out)
    bs = fx.Proxy(graph_out.placeholder('b', torch.Tensor), tracer_out)

    if f_in is None:
        size = x.shape[:-1]
        outsize = size + (irreps_out.dim,)
    else:
        size = x.shape[:-2]
        outsize = size + (f_out, irreps_out.dim,)

    bias_numel = sum(irreps_out[i.i_out].dim for i in instructions if i.i_in == -1)
    if bias_numel > 0:
        if f_out is None:
            bs = bs.reshape(-1, bias_numel)
        else:
            bs = bs.reshape(-1, f_out, bias_numel)

    # = Short-circut for nothing to do =
    # We produce no code for empty instructions
    instructions = [ins for ins in instructions if 0 not in ins.path_shape]

    if len(instructions) == 0 and bias_numel == 0:
        out = x.new_zeros(outsize)

        graph_out.output(out.node, torch.Tensor)
        # Short circut
        # 0 is weight_numel
        return fx.GraphModule({}, graph_out, "linear_forward"), 0, 0

    if f_in is None:
        x = x.reshape(-1, irreps_in.dim)
    else:
        x = x.reshape(-1, f_in, irreps_in.dim)
    batch_out = x.shape[0]

    weight_numel = sum(prod(ins.path_shape) for ins in instructions if ins.i_in != -1)
    if weight_numel > 0:
        ws = ws.reshape(-1, weight_numel) if f_in is None else ws.reshape(-1, f_in, f_out, weight_numel)

    # = extract individual input irreps =
    if len(irreps_in) == 1:
        x_list = [x.reshape(batch_out, *(() if f_in is None else (f_in,)), irreps_in[0].mul, irreps_in[0].ir.dim)]
    else:
        x_list = [
            x.narrow(-1, i.start, mul_ir.dim).reshape(batch_out, *(() if f_in is None else (f_in,)), mul_ir.mul, mul_ir.ir.dim)
            for i, mul_ir in zip(irreps_in.slices(), irreps_in)
        ]

    z = '' if shared_weights else 'z'

    flat_weight_index = 0
    flat_bias_index = 0

    out_list = []

    for ins in instructions:
        mul_ir_out = irreps_out[ins.i_out]

        if ins.i_in == -1:
            # = bias =
            b = bs.narrow(-1, flat_bias_index, prod(ins.path_shape))
            flat_bias_index += prod(ins.path_shape)
            out_list += [(ins.path_weight * b).reshape(1, *(() if f_out is None else (f_out,)), mul_ir_out.dim)]
        else:
            mul_ir_in = irreps_in[ins.i_in]

            # Short-circut for empty irreps
            if mul_ir_in.dim == 0 or mul_ir_out.dim == 0:
                continue

            # Extract the weight from the flattened weight tensor
            path_nweight = prod(ins.path_shape)
            if len(instructions) == 1:
                # Avoid unnecessary view when there is only one weight
                w = ws
            else:
                w = ws.narrow(-1, flat_weight_index, path_nweight)
            w = w.reshape(
                (() if shared_weights else (-1,)) + (() if f_in is None else (f_in, f_out)) + ins.path_shape
            )
            flat_weight_index += path_nweight

            if f_in is None:
                ein_out = torch.einsum(f"{z}uw,zui->zwi", w, x_list[ins.i_in])
            else:
                ein_out = torch.einsum(f"{z}xyuw,zxui->zywi", w, x_list[ins.i_in])

            ein_out = ins.path_weight * ein_out

            out_list += [ein_out.reshape(batch_out, *(() if f_out is None else (f_out,)), mul_ir_out.dim)]

    # = Return the result =
    out = [
        _sum_tensors(
            [out for ins, out in zip(instructions, out_list) if ins.i_out == i_out],
            shape=(batch_out, *(() if f_out is None else (f_out,)), mul_ir_out.dim),
            like=x
        )
        for i_out, mul_ir_out in enumerate(irreps_out)
        if mul_ir_out.mul > 0
    ]
    if len(out) > 1:
        out = torch.cat(out, dim=-1)
    else:
        out = out[0]

    out = out.reshape(outsize)

    graph_out.output(out.node, torch.Tensor)

    # check graphs
    graph_out.lint()

    graphmod_out = fx.GraphModule({}, graph_out, "linear_forward")

    # TODO: when eliminate_dead_code() is in PyTorch stable, use that
    if optimize_einsums:
        # See _tensor_product/_codegen.py for notes
        batchdim = 4
        example_inputs = (
            torch.zeros((batchdim, *(() if f_in is None else (f_in,)), irreps_in.dim)),
            torch.zeros(
                1 if shared_weights else batchdim,
                f_in or 1,
                f_out or 1,
                weight_numel,
            ),
            torch.zeros(
                1 if shared_weights else batchdim,
                f_out or 1,
                bias_numel,
            ),
        )
        graphmod_out = optimize_einsums_full(graphmod_out, example_inputs)

    return graphmod_out, weight_numel, bias_numel



###########################################################################################
# Higher Order Real Clebsch Gordan (based on e3nn by Mario Geiger)
# Authors: Ilyes Batatia
# This program is distributed under the MIT License (see MIT.md)
###########################################################################################

import collections
import itertools
import os
from typing import Iterator, List, Union

import numpy as np
import torch
from macer.externals.e3nn_legacy import o3

try:
    import cuequivariance as cue

    CUET_AVAILABLE = True
except ImportError:
    CUET_AVAILABLE = False

USE_CUEQ_CG = os.environ.get("MACE_USE_CUEQ_CG", "0").lower() in (
    "1",
    "true",
    "yes",
    "y",
)

_TP = collections.namedtuple("_TP", "op, args")
_INPUT = collections.namedtuple("_INPUT", "tensor, start, stop")


def _wigner_nj(
    irrepss: List[o3.Irreps],
    normalization: str = "component",
    filter_ir_mid=None,
    dtype=None,
):
    irrepss = [o3.Irreps(irreps) for irreps in irrepss]
    if filter_ir_mid is not None:
        filter_ir_mid = [o3.Irrep(ir) for ir in filter_ir_mid]

    if len(irrepss) == 1:
        (irreps,) = irrepss
        ret = []
        e = torch.eye(irreps.dim, dtype=dtype)
        i = 0
        for mul, ir in irreps:
            for _ in range(mul):
                sl = slice(i, i + ir.dim)
                ret += [(ir, _INPUT(0, sl.start, sl.stop), e[sl])]
                i += ir.dim
        return ret

    *irrepss_left, irreps_right = irrepss
    ret = []
    for ir_left, path_left, C_left in _wigner_nj(
        irrepss_left,
        normalization=normalization,
        filter_ir_mid=filter_ir_mid,
        dtype=dtype,
    ):
        i = 0
        for mul, ir in irreps_right:
            for ir_out in ir_left * ir:
                if filter_ir_mid is not None and ir_out not in filter_ir_mid:
                    continue

                C = o3.wigner_3j(ir_out.l, ir_left.l, ir.l, dtype=dtype)
                if normalization == "component":
                    C *= ir_out.dim**0.5
                if normalization == "norm":
                    C *= ir_left.dim**0.5 * ir.dim**0.5

                C = torch.einsum("jk,ijl->ikl", C_left.flatten(1), C)
                C = C.reshape(
                    ir_out.dim, *(irreps.dim for irreps in irrepss_left), ir.dim
                )
                for u in range(mul):
                    E = torch.zeros(
                        ir_out.dim,
                        *(irreps.dim for irreps in irrepss_left),
                        irreps_right.dim,
                        dtype=dtype,
                    )
                    sl = slice(i + u * ir.dim, i + (u + 1) * ir.dim)
                    E[..., sl] = C
                    ret += [
                        (
                            ir_out,
                            _TP(
                                op=(ir_left, ir, ir_out),
                                args=(
                                    path_left,
                                    _INPUT(len(irrepss_left), sl.start, sl.stop),
                                ),
                            ),
                            E,
                        )
                    ]
            i += mul * ir.dim
    return sorted(ret, key=lambda x: x[0])


def U_matrix_real(
    irreps_in: Union[str, o3.Irreps],
    irreps_out: Union[str, o3.Irreps],
    correlation: int,
    normalization: str = "component",
    filter_ir_mid=None,
    dtype=None,
    use_cueq_cg=True,
    use_nonsymmetric_product=False,
):
    irreps_out = o3.Irreps(irreps_out)
    irrepss = [o3.Irreps(irreps_in)] * correlation

    if use_cueq_cg is None:
        use_cueq_cg = USE_CUEQ_CG
    if correlation == 4 and not use_cueq_cg:
        filter_ir_mid = [(i, 1 if i % 2 == 0 else -1) for i in range(12)]
    if use_cueq_cg and CUET_AVAILABLE:
        return compute_U_cueq(
            irreps_in, irreps_out=irreps_out, correlation=correlation, dtype=dtype
        )

    try:
        wigners = _wigner_nj(irrepss, normalization, filter_ir_mid, dtype)
    except NotImplementedError as e:
        if CUET_AVAILABLE:
            return compute_U_cueq(
                irreps_in,
                irreps_out=irreps_out,
                correlation=correlation,
                use_nonsymmetric_product=use_nonsymmetric_product,
                dtype=dtype,
            )
        raise NotImplementedError(
            "The requested Clebsch-Gordan coefficients are not implemented, please install cuequivariance; pip install cuequivariance"
        ) from e

    current_ir = wigners[0][0]
    out = []
    stack = torch.tensor([])

    for ir, _, base_o3 in wigners:
        if ir in irreps_out and ir == current_ir:
            stack = torch.cat((stack, base_o3.squeeze().unsqueeze(-1)), dim=-1)
            last_ir = current_ir
        elif ir in irreps_out and ir != current_ir:
            if len(stack) != 0:
                out += [last_ir, stack]
            stack = base_o3.squeeze().unsqueeze(-1)
            current_ir, last_ir = ir, ir
        else:
            current_ir = ir
    try:
        out += [last_ir, stack]
    except:  # pylint: disable=bare-except
        first_dim = irreps_out.dim
        if first_dim != 1:
            size = [first_dim] + [o3.Irreps(irreps_in).dim] * correlation + [1]
        else:
            size = [o3.Irreps(irreps_in).dim] * correlation + [1]
        out = [str(irreps_out)[:-2], torch.zeros(size, dtype=dtype)]
    return out


if CUET_AVAILABLE:

    def compute_U_cueq(
        irreps_in, irreps_out, correlation=2, use_nonsymmetric_product=False, dtype=None
    ):
        if dtype is None:
            dtype = torch.get_default_dtype()
        U = []
        irreps_in = cue.Irreps(O3_e3nn, str(irreps_in))
        irreps_out = cue.Irreps(O3_e3nn, str(irreps_out))
        for _, ir in irreps_out:
            try:
                U_matrix_full_symm = cue.reduced_symmetric_tensor_product_basis(
                    irreps_in,
                    correlation,
                    keep_ir=ir,
                    layout=cue.ir_mul,
                )
                U_matrix_full_symm = U_matrix_full_symm.array
                if use_nonsymmetric_product:
                    try:
                        U_matrix_full_antisymmetric = (
                            cue.reduced_antisymmetric_tensor_product_basis(
                                irreps_in,
                                correlation,
                                keep_ir=ir,
                                layout=cue.ir_mul,
                            ).array
                        )
                        U_matrix_full = torch.cat(
                            (U_matrix_full_symm, U_matrix_full_antisymmetric), dim=-1
                        )
                    except ValueError:
                        continue
                else:
                    U_matrix_full = U_matrix_full_symm

            except ValueError:
                if ir.dim == 1:
                    out_shape = (*([irreps_in.dim] * correlation), 1)
                else:
                    out_shape = (ir.dim, *([irreps_in.dim] * correlation), 1)
                return [
                    torch.zeros(
                        out_shape,
                        dtype=torch.get_default_dtype(),
                    )
                ]
            if U_matrix_full.shape[-1] == 0:
                if ir.dim == 1:
                    out_shape = (*([irreps_in.dim] * correlation), 1)
                else:
                    out_shape = (ir.dim, *([irreps_in.dim] * correlation), 1)
                return [
                    torch.zeros(
                        out_shape,
                        dtype=torch.get_default_dtype(),
                    )
                ]
            ir_str = str(ir)
            U.append(ir_str)
            U_matrix_full = torch.tensor(
                U_matrix_full.reshape(*([irreps_in.dim] * correlation), ir.dim, -1),
                dtype=dtype,
            )
            U_matrix_full = torch.moveaxis(U_matrix_full, -2, 0)
            if ir.dim == 1:
                U_matrix_full = U_matrix_full[0]
            U.append(U_matrix_full)
        return U

    class O3_e3nn(cue.O3):
        def __mul__(  # pylint: disable=no-self-argument
            rep1: "O3_e3nn", rep2: "O3_e3nn"
        ) -> Iterator["O3_e3nn"]:
            return [O3_e3nn(l=ir.l, p=ir.p) for ir in cue.O3.__mul__(rep1, rep2)]

        @classmethod
        def clebsch_gordan(
            cls, rep1: "O3_e3nn", rep2: "O3_e3nn", rep3: "O3_e3nn"
        ) -> np.ndarray:
            rep1, rep2, rep3 = cls._from(rep1), cls._from(rep2), cls._from(rep3)

            if rep1.p * rep2.p == rep3.p:
                return o3.wigner_3j(rep1.l, rep2.l, rep3.l).numpy()[None] * np.sqrt(
                    rep3.dim
                )
            return np.zeros((0, rep1.dim, rep2.dim, rep3.dim))

        def __lt__(  # pylint: disable=no-self-argument
            rep1: "O3_e3nn", rep2: "O3_e3nn"
        ) -> bool:
            rep2 = rep1._from(rep2)
            return (rep1.l, rep1.p) < (rep2.l, rep2.p)

        @classmethod
        def iterator(cls) -> Iterator["O3_e3nn"]:
            for l in itertools.count(0):
                yield O3_e3nn(l=l, p=1 * (-1) ** l)
                yield O3_e3nn(l=l, p=-1 * (-1) ** l)

else:

    class O3_e3nn:
        pass




import torch
import sys
import os

# --- MPS Compatibility Patch ---
# MACE internally calls .double() which fails on MPS.
# We monkey-patch torch.Tensor.double to return float32 if the tensor is on MPS.
orig_double = torch.Tensor.double

def mps_safe_double(self):
    if self.device.type == 'mps':
        # Return float32 instead of crashing on MPS
        return self.float()
    return orig_double(self)

torch.Tensor.double = mps_safe_double
# Also patch to() for explicit double conversions
orig_to = torch.Tensor.to
def mps_safe_to(self, *args, **kwargs):
    if 'dtype' in kwargs and kwargs['dtype'] == torch.float64 and self.device.type == 'mps':
        kwargs['dtype'] = torch.float32
    if len(args) > 0 and args[0] == torch.float64 and self.device.type == 'mps':
        args = (torch.float32,) + args[1:]
    return orig_to(self, *args, **kwargs)
torch.Tensor.to = mps_safe_to

print("INFO: Applied MPS float64 -> float32 monkey-patch for MACE.")

# --- Execute MACE ---
# We try to import from the path provided in the environment or fallback to standard
try:
    # Use the module name passed via arguments
    module_name = sys.argv[1]
    import importlib
    mace_cli = importlib.import_module(module_name)
    # Remove the wrapper-specific argument before passing to MACE
    sys.argv.pop(1)
    mace_cli.main()
except Exception as e:
    print(f"ERROR in MPS Wrapper: {e}")
    sys.exit(1)



"""Module to compute the Brillouin zone of a crystal."""

from collections import defaultdict
import warnings
from typing import Union

import numpy as np
from scipy.spatial import Voronoi, ConvexHull, Delaunay


def get_BZ(
    b1: Union[list, np.array], b2: Union[list, np.array], b3: Union[list, np.array]
) -> dict:
    """Get the faces of the BZ given the three vectors b1,b2,b3.

    Args:
        b1 (Union[list, np.array]): First vector of the reciprocal lattice.
        b2 (Union[list, np.array]): Second vector of the reciprocal lattice.
        b3 (Union[list, np.array]): Third vector of the reciprocal lattice.

    Returns:
        dict: Both triangular faces (oriented) and flat faces (if your
            plotting library prefers these - these are not oriented for the
            time being)
    """
    warnings.warn(
        '`get_BZ` is deprecated, use the `BZ` class instead, to represent a brillouine zone. '
        'The faces and triangles can be accesses as attributes.',
        UserWarning,
    )

    bz = BZ(b1, b2, b3)

    faces_data = {
        'triangles_vertices': bz.triangles_vertices,
        'triangles': bz.triangles,
        'faces': bz.faces,
    }

    return faces_data


class BZ:
    """Class to compute the Brillouin zone of a crystal."""

    def __init__(
        self,
        b1: Union[list, np.array],
        b2: Union[list, np.array],
        b3: Union[list, np.array],
    ) -> None:
        """Initialize the Brillouin zone."""
        self._b_vectors = (b1, b2, b3)
        self._initialize_BZ(b1, b2, b3)

    @property
    def b_vectors(self) -> tuple:
        """Return the reciprocal lattice vectors."""
        return self._b_vectors

    @property
    def hull(self) -> ConvexHull:
        """Return the convex hull of the Brillouin zone."""
        return self._hull

    @property
    def delaunay(self) -> Delaunay:
        """Return the Delaunay triangulation of the Brillouin zone."""
        return self._delaunay

    @property
    def triangles_vertices(self) -> Union[list, np.array]:
        """Return the vertices of the triangles of the Brillouin zone."""
        return self._triangles_vertices

    @property
    def triangles(self) -> Union[list, np.array]:
        """Return the triangles of the Brillouin zone."""
        return self._triangles

    @property
    def faces(self) -> Union[list, np.array]:
        """Return the faces of the Brillouin zone."""
        return self._faces

    def _get_missing_point(self, tr: Union[list, np.array], p1: int, p2: int) -> int:
        """
        tr is a list of 3 indices (the indices of the vertices of a given triangle).
        p1 and p2 must be in the tr list, the third index is returned.

        ValueError is raised if more than one entry is not among p1 and p2,
        or if all points are among p1 and p2 (e.g. a point repeated twice)
        """
        missing = None
        for vertex in tr:
            if vertex not in (p1, p2):
                if missing is not None:
                    raise ValueError('Two missing points found...')
                missing = vertex

        if missing is None:
            raise ValueError('No missing points!')

        return missing

    def _are_coplanar(
        self,
        v1: Union[list, np.array],
        v2: Union[list, np.array],
        v3: Union[list, np.array],
    ) -> bool:
        """
        v1, v2, v3: 3D vectors.
        Return True if they are coplanar, False otherwise
        """
        return float(abs(np.dot(np.cross(v1, v2), v3))) < 1.0e-6

    def _initialize_BZ(
        self,
        b1: Union[list, np.array],
        b2: Union[list, np.array],
        b3: Union[list, np.array],
    ) -> None:
        """
        Get the faces of the BZ given the three vectors b1,b2,b3.
        Return both triangular faces (oriented) and flat faces (if your
            plotting library prefers these - these are not oriented for the
            time being)
        """
        ret_data = {}

        supercell_size = 3  # Is this enough?

        # G-vectors
        points3d = []
        central_idx = None
        for i in range(-supercell_size, supercell_size + 1):
            for j in range(-supercell_size, supercell_size + 1):
                for k in range(-supercell_size, supercell_size + 1):
                    if i == 0 and j == 0 and k == 0:
                        central_idx = len(points3d)
                    points3d.append(
                        i * np.array(b1) + j * np.array(b2) + k * np.array(b3)
                    )

        # Get Voronois
        vor3d = Voronoi(np.array(points3d))
        # Get the vertices of the central" Voronoi( around the origin G=0)
        central_voronoi_3d = np.array(
            [
                vor3d.vertices[idx]
                for idx in vor3d.regions[vor3d.point_region[central_idx]]
            ]
        )

        # Get the convex hull of these points (all triangular faces)
        hull = ConvexHull(central_voronoi_3d)

        ## REORIENT TRIANGLES
        ## NOTE! TODO: I should do the same for faces
        ret_data['triangles_vertices'] = hull.points.tolist()
        ## Naive one
        # ret_data['triangles'] = hull.simplices.tolist()
        ## Instead, I orient them all
        ret_data['triangles'] = []
        for simplex in hull.simplices:
            points = np.array([hull.points[i] for i in simplex])
            center = points.sum(axis=0) / float(len(points))

            # Get normal vector (with direction!)
            normal = np.cross(center - points[1], center - points[0])
            # Normalize, then rescale to a small value
            normal = normal / np.linalg.norm(normal)
            max_length = np.sqrt((points**2).sum(axis=1).max())
            normal /= max_length
            normal *= 1.0e-4
            point_up = center + normal
            point_down = center - normal
            delaunay = Delaunay(hull.points)
            is_up_inside = delaunay.find_simplex(point_up) >= 0
            is_down_inside = delaunay.find_simplex(point_down) >= 0

            if is_up_inside and not is_down_inside:
                correct_orientation = True
            elif not is_up_inside and is_down_inside:
                correct_orientation = False
            else:
                inside_outside_string = 'inside' if is_up_inside else 'outside'
                print(
                    f'WARNING! Both vectors are {inside_outside_string}... not changing orientation'
                )
                correct_orientation = True

            if correct_orientation:
                ret_data['triangles'].append(simplex.tolist())
            else:
                ret_data['triangles'].append(simplex[::-1].tolist())

        # print hull.area, hull.volume

        # Get edge-sharing faces
        # edges has as key a tuple (sorted) with the indices of the two vertices of
        # the shared edge; the value are the indices of the triangles
        edges = defaultdict(list)
        for simplex_idx, simplex in enumerate(hull.simplices):
            edges[tuple(sorted([simplex[0], simplex[1]]))].append(simplex_idx)
            edges[tuple(sorted([simplex[1], simplex[2]]))].append(simplex_idx)
            edges[tuple(sorted([simplex[2], simplex[0]]))].append(simplex_idx)
        # convert to dictionary of lists (from defaultdict of sets)
        edges = dict(edges)

        ### Create now the list of faces, merging the triangles that share an
        ### edge and are coplanar. Note: this works only if up to two triangles
        ### must be merged; if three or more, this will not produce the expected
        ### result
        # print edges

        # Store merge operations to perform
        merge_with = defaultdict(set)

        # List of found simplices that have been merged; will be used at the
        # end to add the triangles that have not been merged, if any
        # TODO (was not commented out before but never used)
        # merged_simplices = []
        for (p1, p2), triangles in edges.items():
            # I do it many times, but it's the easiest way (and anyway it's
            # a set, so it should be fast: I add a point to be merged with
            # itself
            merge_with[triangles[0]].add(triangles[0])
            merge_with[triangles[1]].add(triangles[1])

            if len(triangles) != 2:
                # An edge shared by less (or more) than 2 triangles?
                print('Warning!', p1, p2, triangles)
            else:
                # Check if two triangles are coplanar: get the other two
                # vertices that are not on the shared edge
                otherpoint0 = self._get_missing_point(
                    hull.simplices[triangles[0]], p1, p2
                )
                otherpoint1 = self._get_missing_point(
                    hull.simplices[triangles[1]], p1, p2
                )
                # The actual vector coordinates
                otherpoint0_p = hull.points[otherpoint0]
                otherpoint1_p = hull.points[otherpoint1]
                p1_p = hull.points[p1]
                p2_p = hull.points[p2]

                # Check if they are coplanar
                if self._are_coplanar(
                    p2_p - p1_p, otherpoint0_p - p1_p, otherpoint1_p - p1_p
                ):
                    merge_with[triangles[0]].add(triangles[1])
                    merge_with[triangles[1]].add(triangles[0])

        # PROBLEM TO SOLVE: we have to put together all groups
        ## E.g. we have now:
        # 0: [0, 3]
        # 1: [1, 2, 3]
        # 2: [1, 2]
        # 3: [0, 1, 3]
        ## We should get instead for all [0,1,2,3]
        # So we have to do a pass to merge them all
        # The algorithm below is probably wrong (actually, it is only
        # probably inefficient)

        # for k, v in merge_with.items():
        #    print "{}: {}".format(k, list(v))

        # Iterate untile convergence - not sure this is correct
        has_changed = True
        while has_changed:
            has_changed = False
            # Add the missing ones
            for tr in range(len(hull.simplices)):
                for other1 in merge_with[tr]:
                    for other2 in merge_with[tr]:
                        if other1 not in merge_with[other2]:
                            has_changed = True
                            merge_with[other2].add(other1)
                        if other2 not in merge_with[other1]:
                            has_changed = True
                            merge_with[other1].add(other2)

        # convert to dict, and most importantly convert to list and sort
        merge_with = {k: sorted(v) for k, v in merge_with.items()}

        # Assign to the smallest integer idx
        merge_group = {k: v[0] for k, v in merge_with.items()}

        groups = defaultdict(list)
        # I create a reverse index
        for k, v in merge_group.items():
            groups[v].append(k)

        # List of faces (elements are lists of vertex ids)
        faces = []

        for group in groups.values():
            if len(group) == 1:
                faces.append(
                    [hull.points[point_idx] for point_idx in hull.simplices[group[0]]]
                )
            else:
                # Get all points
                all_points_idx = sorted(
                    set(np.concatenate([hull.simplices[g] for g in group]))
                )

                all_points_coords = [
                    hull.points[point_idx] for point_idx in all_points_idx
                ]
                # Find projection in 2D: I first choose a first vector (between
                # two points v1; I find the orthogonal vector to the plane b;
                # I find a second vector v2 orthogonal to v1 and on the plane
                # (<=> orthogonal to v1 and b); I normalize v1 and v2;
                # I get the components of the vectors w.r.t. v1 and v2
                # NOTE: there is at least 1 triangle => at least 3 points
                v1 = all_points_coords[1] - all_points_coords[0]
                temp_v2 = all_points_coords[2] - all_points_coords[0]
                b = np.cross(v1, temp_v2)
                v2 = np.cross(v1, b)
                # Normalize
                v1 = v1 / np.linalg.norm(v1)
                v2 = v2 / np.linalg.norm(v2)
                # get components
                x = [np.dot(point, v1) for point in all_points_coords]
                y = [np.dot(point, v2) for point in all_points_coords]
                # 2. do convexhull in 2D
                hull_face2d = ConvexHull(np.array([x, y]).T)
                # 3. get point indices of convex hull, convert back to original
                # 3D indices
                # 2dhull.vertices contains the segments, but with ids in the
                # smaller subset. We want the indices in the initial set:
                actual_points_idx = [
                    all_points_idx[subset_idx] for subset_idx in hull_face2d.vertices
                ]
                # 4. add to faces list
                faces.append(
                    [hull.points[point_idx].tolist() for point_idx in actual_points_idx]
                )

        ret_data['faces'] = faces

        self._hull = hull
        self._delaunay = Delaunay(hull.points)

        self._triangles_vertices = ret_data['triangles_vertices']
        self._triangles = ret_data['triangles']
        self._faces = ret_data['faces']

    def is_inside_bz(self, p: Union[list, np.array]) -> bool:
        """Check if a point is within the convex hull of the Brillouin zone.

        Args:
            p (Union[list, np.array]): Point to check.

        Returns:
            bool: True if the point is inside the Brillouin zone, False otherwise.
        """
        # returns -1 if no solution is found
        return self._delaunay.find_simplex(p) >= 0


if __name__ == '__main__':
    import matplotlib.pyplot as plt  # pylint: disable=import-error
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection  # pylint: disable=import-error

    # draw a vector
    from matplotlib.patches import FancyArrowPatch  # pylint: disable=import-error
    from mpl_toolkits.mplot3d import proj3d  # pylint: disable=import-error

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def do_3d_projection(self, renderer=None):  # pylint: disable=unused-argument
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            return np.min(zs)

    ##SC
    # faces_data = get_BZ(b1 = [1,0,0], b2 = [0,1,0], b3 = [0,0,1])
    ##BCC
    # faces_data = get_BZ(b1 = [1,1,0], b2 = [1,0,1], b3 = [0,1,1])
    ##FCC
    bz = BZ(b1=[1, 1, -1], b2=[1, -1, 1], b3=[-1, 1, 1])

    faces_coords = bz.faces

    faces_count = defaultdict(int)
    for face in faces_coords:
        faces_count[len(face)] += 1

    for num_sides in sorted(faces_count.keys()):
        print(f'{num_sides} faces: {faces_count[num_sides]}')

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.add_collection(
        Poly3DCollection(
            faces_coords, linewidth=1, alpha=0.9, edgecolor='k', facecolor='#ccccff'
        )
    )

    # draw origin
    ax.scatter([0], [0], [0], color='g', s=100)

    axes_length = 2
    # Add axes
    ax.add_artist(
        Arrow3D(
            (0, axes_length),
            (0, 0),
            (0, 0),
            mutation_scale=20,
            lw=1,
            arrowstyle='-|>',
            color='k',
        )
    )
    ax.add_artist(
        Arrow3D(
            (0, 0),
            (0, axes_length),
            (0, 0),
            mutation_scale=20,
            lw=1,
            arrowstyle='-|>',
            color='k',
        )
    )
    ax.add_artist(
        Arrow3D(
            (0, 0),
            (0, 0),
            (0, axes_length),
            mutation_scale=20,
            lw=1,
            arrowstyle='-|>',
            color='k',
        )
    )

    ## Reset limits
    ax.set_xlim(-1, 1)
    ax.set_ylim(-1, 1)
    ax.set_zlim(-1, 1)
    ax.axis('off')
    ax.view_init(elev=0, azim=60)

    plt.show()
